"""
Monthly values
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2020 Project Author Pilar Monsalvete pilar_monsalvete@yahoo.es
"""

import pandas as pd
import numpy as np
import calendar as cal


class MonthlyValues:
  def __init__(self):
    self._month_hour = None

  def get_mean_values(self, values):
    out = None
    if values is not None:
      if 'month' not in values.columns:
        values = pd.concat([self.month_hour, pd.DataFrame(values)], axis=1)
      out = values.groupby('month', as_index=False).mean()
      del out['month']
    return out

  def get_total_month(self, values):
    out = None
    if values is not None:
      if 'month' not in values.columns:
        values = pd.concat([self.month_hour, pd.DataFrame(values)], axis=1)
      out = pd.DataFrame(values).groupby('month', as_index=False).sum()
      del out['month']
    return out

  @property
  def month_hour(self):
    """
    returns a DataFrame that has x values of the month number (January = 1, February = 2...),
    being x the number of hours of the corresponding month
    :return: DataFrame(int)
    """
    array = []
    for i in range(0, 12):
      days_of_month = cal.monthrange(2015, i+1)[1]
      total_hours = days_of_month * 24
      array = np.concatenate((array, np.full(total_hours, i + 1)))
    self._month_hour = pd.DataFrame(array, columns=['month'])
    return self._month_hour



"""
ConstructionFactory (before PhysicsFactory) retrieve the specific construction module for the given region
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Atiya atiya.atiya@mail.concordia.ca
"""

from pathlib import Path
from hub.imports.life_cycle_assessment.lca_fuel import LcaFuel
from hub.imports.life_cycle_assessment.lca_vehicle import LcaVehicle
from hub.imports.life_cycle_assessment.lca_machine import LcaMachine
from hub.imports.life_cycle_assessment.lca_material import LcaMaterial
from hub.helpers.utils import validate_import_export_type
from hub.hub_logger import logger


class LifeCycleAssessment:
  """
  Life cycle assessment factory class
  """
  def __init__(self, handler, city, base_path=None):
    if base_path is None:
      base_path = Path(Path(__file__).parent.parent / 'data/life_cycle_assessment')
    self._handler = '_' + handler.lower().replace(' ', '_')
    class_funcs = validate_import_export_type(LifeCycleAssessment)
    if self._handler not in class_funcs:
      err_msg = f"Wrong import type. Valid functions include {class_funcs}"
      logger.error(err_msg)
      raise Exception(err_msg)
    self._city = city
    self._base_path = base_path

  def _fuel(self):
    """
    Enrich the city by adding the fuel carbon information
    """
    LcaFuel(self._city, self._base_path).enrich()

  def _vehicle(self):
    """
    Enrich the city by adding the vehicle carbon information
    """
    LcaVehicle(self._city, self._base_path).enrich()

  def _machine(self):
    """
    Enrich the city by adding the machine carbon information
    """
    LcaMachine(self._city, self._base_path).enrich()

  def _material(self):
    """
    Enrich the city by adding the material carbon information
    """
    LcaMaterial(self._city, self._base_path).enrich()

  def enrich(self):
    """
    Enrich the city given to the class using the class given handler
    :return: None
    """
    getattr(self, self._handler, lambda: None)()



"""
Dictionaries module for Pluto function to hub function
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import hub.helpers.constants as cte


class PlutoFunctionToHubFunction:

  def __init__(self):
    self._dictionary = {
    'A0': cte.SINGLE_FAMILY_HOUSE,
    'A1': cte.SINGLE_FAMILY_HOUSE,
    'A2': cte.SINGLE_FAMILY_HOUSE,
    'A3': cte.SINGLE_FAMILY_HOUSE,
    'A4': cte.SINGLE_FAMILY_HOUSE,
    'A5': cte.SINGLE_FAMILY_HOUSE,
    'A6': cte.SINGLE_FAMILY_HOUSE,
    'A7': cte.SINGLE_FAMILY_HOUSE,
    'A8': cte.SINGLE_FAMILY_HOUSE,
    'A9': cte.SINGLE_FAMILY_HOUSE,
    'B1': cte.MULTI_FAMILY_HOUSE,
    'B2': cte.MULTI_FAMILY_HOUSE,
    'B3': cte.MULTI_FAMILY_HOUSE,
    'B9': cte.MULTI_FAMILY_HOUSE,
    'C0': cte.RESIDENTIAL,
    'C1': cte.RESIDENTIAL,
    'C2': cte.RESIDENTIAL,
    'C3': cte.RESIDENTIAL,
    'C4': cte.RESIDENTIAL,
    'C5': cte.RESIDENTIAL,
    'C6': cte.RESIDENTIAL,
    'C7': cte.RESIDENTIAL,
    'C8': cte.RESIDENTIAL,
    'C9': cte.RESIDENTIAL,
    'D0': cte.RESIDENTIAL,
    'D1': cte.RESIDENTIAL,
    'D2': cte.RESIDENTIAL,
    'D3': cte.RESIDENTIAL,
    'D4': cte.RESIDENTIAL,
    'D5': cte.RESIDENTIAL,
    'D6': cte.RESIDENTIAL,
    'D7': cte.RESIDENTIAL,
    'D8': cte.RESIDENTIAL,
    'D9': cte.RESIDENTIAL,
    'E1': cte.WAREHOUSE,
    'E3': cte.WAREHOUSE,
    'E4': cte.WAREHOUSE,
    'E5': cte.WAREHOUSE,
    'E7': cte.WAREHOUSE,
    'E9': cte.WAREHOUSE,
    'F1': cte.WAREHOUSE,
    'F2': cte.WAREHOUSE,
    'F4': cte.WAREHOUSE,
    'F5': cte.WAREHOUSE,
    'F8': cte.WAREHOUSE,
    'F9': cte.WAREHOUSE,
    'G0': cte.SMALL_OFFICE,
    'G1': cte.SMALL_OFFICE,
    'G2': cte.SMALL_OFFICE,
    'G3': cte.SMALL_OFFICE,
    'G4': cte.SMALL_OFFICE,
    'G5': cte.SMALL_OFFICE,
    'G6': cte.SMALL_OFFICE,
    'G7': cte.SMALL_OFFICE,
    'G8': cte.SMALL_OFFICE,
    'G9': cte.SMALL_OFFICE,
    'H1': cte.HOTEL,
    'H2': cte.HOTEL,
    'H3': cte.HOTEL,
    'H4': cte.HOTEL,
    'H5': cte.HOTEL,
    'H6': cte.HOTEL,
    'H7': cte.HOTEL,
    'H8': cte.HOTEL,
    'H9': cte.HOTEL,
    'HB': cte.HOTEL,
    'HH': cte.HOTEL,
    'HR': cte.HOTEL,
    'HS': cte.HOTEL,
    'I1': cte.HOSPITAL,
    'I2': cte.OUT_PATIENT_HEALTH_CARE,
    'I3': cte.OUT_PATIENT_HEALTH_CARE,
    'I4': cte.RESIDENTIAL,
    'I5': cte.OUT_PATIENT_HEALTH_CARE,
    'I6': cte.OUT_PATIENT_HEALTH_CARE,
    'I7': cte.OUT_PATIENT_HEALTH_CARE,
    'I9': cte.OUT_PATIENT_HEALTH_CARE,
    'J1': cte.LARGE_OFFICE,
    'J2': cte.LARGE_OFFICE,
    'J3': cte.LARGE_OFFICE,
    'J4': cte.LARGE_OFFICE,
    'J5': cte.LARGE_OFFICE,
    'J6': cte.LARGE_OFFICE,
    'J7': cte.LARGE_OFFICE,
    'J8': cte.LARGE_OFFICE,
    'J9': cte.LARGE_OFFICE,
    'K1': cte.STRIP_MALL,
    'K2': cte.STRIP_MALL,
    'K3': cte.STRIP_MALL,
    'K4': cte.RESIDENTIAL,
    'K5': cte.RESTAURANT,
    'K6': cte.SUPERMARKET,
    'K7': cte.SUPERMARKET,
    'K8': cte.SUPERMARKET,
    'K9': cte.SUPERMARKET,
    'L1': cte.RESIDENTIAL,
    'L2': cte.RESIDENTIAL,
    'L3': cte.RESIDENTIAL,
    'L8': cte.RESIDENTIAL,
    'L9': cte.RESIDENTIAL,
    'M1': cte.LARGE_OFFICE,
    'M2': cte.LARGE_OFFICE,
    'M3': cte.LARGE_OFFICE,
    'M4': cte.LARGE_OFFICE,
    'M9': cte.LARGE_OFFICE,
    'N1': cte.RESIDENTIAL,
    'N2': cte.RESIDENTIAL,
    'N3': cte.RESIDENTIAL,
    'N4': cte.RESIDENTIAL,
    'N9': cte.RESIDENTIAL,
    'O1': cte.SMALL_OFFICE,
    'O2': cte.SMALL_OFFICE,
    'O3': cte.SMALL_OFFICE,
    'O4': cte.SMALL_OFFICE,
    'O5': cte.SMALL_OFFICE,
    'O6': cte.SMALL_OFFICE,
    'O7': cte.SMALL_OFFICE,
    'O8': cte.SMALL_OFFICE,
    'O9': cte.SMALL_OFFICE,
    'P1': cte.LARGE_OFFICE,
    'P2': cte.HOTEL,
    'P3': cte.SMALL_OFFICE,
    'P4': cte.SMALL_OFFICE,
    'P5': cte.SMALL_OFFICE,
    'P6': cte.SMALL_OFFICE,
    'P7': cte.LARGE_OFFICE,
    'P8': cte.LARGE_OFFICE,
    'P9': cte.SMALL_OFFICE,
    'Q0': cte.SMALL_OFFICE,
    'Q1': cte.SMALL_OFFICE,
    'Q2': cte.SMALL_OFFICE,
    'Q3': cte.SMALL_OFFICE,
    'Q4': cte.SMALL_OFFICE,
    'Q5': cte.SMALL_OFFICE,
    'Q6': cte.SMALL_OFFICE,
    'Q7': cte.SMALL_OFFICE,
    'Q8': cte.SMALL_OFFICE,
    'Q9': cte.SMALL_OFFICE,
    'R0': cte.RESIDENTIAL,
    'R1': cte.RESIDENTIAL,
    'R2': cte.RESIDENTIAL,
    'R3': cte.RESIDENTIAL,
    'R4': cte.RESIDENTIAL,
    'R5': cte.RESIDENTIAL,
    'R6': cte.RESIDENTIAL,
    'R7': cte.RESIDENTIAL,
    'R8': cte.RESIDENTIAL,
    'R9': cte.RESIDENTIAL,
    'RA': cte.RESIDENTIAL,
    'RB': cte.RESIDENTIAL,
    'RC': cte.RESIDENTIAL,
    'RD': cte.RESIDENTIAL,
    'RG': cte.RESIDENTIAL,
    'RH': cte.RESIDENTIAL,
    'RI': cte.RESIDENTIAL,
    'RK': cte.RESIDENTIAL,
    'RM': cte.RESIDENTIAL,
    'RR': cte.RESIDENTIAL,
    'RS': cte.RESIDENTIAL,
    'RW': cte.RESIDENTIAL,
    'RX': cte.RESIDENTIAL,
    'RZ': cte.RESIDENTIAL,
    'S0': cte.RESIDENTIAL,
    'S1': cte.RESIDENTIAL,
    'S2': cte.RESIDENTIAL,
    'S3': cte.RESIDENTIAL,
    'S4': cte.RESIDENTIAL,
    'S5': cte.RESIDENTIAL,
    'S9': cte.RESIDENTIAL,
    'U0': cte.WAREHOUSE,
    'U1': cte.WAREHOUSE,
    'U2': cte.WAREHOUSE,
    'U3': cte.WAREHOUSE,
    'U4': cte.WAREHOUSE,
    'U5': cte.WAREHOUSE,
    'U6': cte.WAREHOUSE,
    'U7': cte.WAREHOUSE,
    'U8': cte.WAREHOUSE,
    'U9': cte.WAREHOUSE,
    'W1': cte.PRIMARY_SCHOOL,
    'W2': cte.PRIMARY_SCHOOL,
    'W3': cte.SECONDARY_SCHOOL,
    'W4': cte.EDUCATION,
    'W5': cte.SECONDARY_SCHOOL,
    'W6': cte.SECONDARY_SCHOOL,
    'W7': cte.SECONDARY_SCHOOL,
    'W8': cte.PRIMARY_SCHOOL,
    'W9': cte.SECONDARY_SCHOOL,
    'Y1': cte.LARGE_OFFICE,
    'Y2': cte.LARGE_OFFICE,
    'Y3': cte.LARGE_OFFICE,
    'Y4': cte.LARGE_OFFICE,
    'Y5': cte.LARGE_OFFICE,
    'Y6': cte.LARGE_OFFICE,
    'Y7': cte.LARGE_OFFICE,
    'Y8': cte.LARGE_OFFICE,
    'Y9': cte.LARGE_OFFICE,
    'Z1': cte.LARGE_OFFICE
  }

  @property
  def dictionary(self) -> dict:
    return self._dictionary



"""
Concordia file report
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

import io
import json
from pathlib import Path
import pandas as pd


class ConcordiaFileReport:
  """
  Concordia file report for sensors base class
  """
  def __init__(self, city, end_point, base_path, db_file):
    self._city_object = []
    self._city_objects_cluster = []
    self._sensors = []
    self._sensor_point = {}
    self._city = city
    self._end_point = end_point
    self._sensor_database = base_path
    metadata = True
    content = False
    with open(Path(base_path / db_file).resolve()) as concordia_db:
      self._sensor_database = json.load(concordia_db)

    for city_object in self._sensor_database['sensors']:
      city_object_name = city_object['city_object']
      for sensor in city_object['sensors']:
        self._city_object.append(city_object_name)
        self._sensors.append(sensor)

    buffer = ""
    with open(end_point.resolve()) as data:
      for line in data:
        line = ConcordiaFileReport._clean_line(line)
        if metadata:
          fields = line.split(',')
          if len(fields) > 2:
            point = fields[0].replace(":", "")
            key = fields[1]
            if fields[1] in self._sensors:
              self._sensor_point[key] = point
        if "End of Report" in line:
          content = False
        if content:
          line = ConcordiaFileReport._merge_date_time(line)
          buffer = buffer + line + '\n'
        if line == '':
          metadata = False
          content = True
    measures = pd.read_csv(io.StringIO(buffer), sep=',')
    measures["Date time"] = pd.to_datetime(measures["Date time"])
    self._measures = ConcordiaFileReport._force_format(measures)

  @staticmethod
  def _clean_line(line):
    return line.replace('"', '').replace('\n', '')

  @staticmethod
  def _merge_date_time(line):
    fields = line.split(',')
    date = fields[0]
    time = fields[1]
    if '<>' in date:
      return line.replace(f'{date},{time}', 'Date time')
    date_fields = date.split('/')
    format_date_time = f'"{int(date_fields[2])}-{int(date_fields[0]):02d}-{int(date_fields[1]):02d} {time}"'
    return line.replace(f'{date},{time}', format_date_time)

  @staticmethod
  def _force_format(df):
    for head in df.head():
      if 'Date time' not in head:
        df = df.astype({head: 'float64'})
    return df



"""
Bus module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from hub.city_model_structure.attributes.schedule import Schedule


class Bus:
  """
  Bus class
  """

  def __init__(self):
    self._maintenance_time = None
    self._charging_time = None
    self._recovery_time = None
    self._vehicle_type = None
    self._energy_consumption = None
    self._trips_schedule = None
    self._capacity = None
    self._maintenance_cost = None
    self._investment_cost = None
    self._charging_range = None
    self._maximum_travel_range = None

  @property
  def maintenance_time(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._maintenance_time

  @property
  def charging_time(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._charging_time

  @property
  def recovery_time(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self.maintenance_time + self.charging_time

  @property
  def vehicle_type(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._vehicle_type

  @property
  def energy_consumption(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._energy_consumption

  @property
  def trips_schedule(self) -> Schedule:
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._trips_schedule

  @property
  def capacity(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._capacity

  @property
  def maintenance_cost(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._maintenance_cost

  @property
  def investment_cost(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._investment_cost

  @property
  def charging_range(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._charging_range

  @property
  def maximum_travel_range(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._maximum_travel_range



"""
CityLayersTest
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder:  Milad Aghamohamadnia --- milad.aghamohamadnia@concordia.ca

"""

from unittest import TestCase
import json
import os
import time
import uuid
from pathlib import Path
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.usage_factory import UsageFactory
from hub.imports.construction_factory import ConstructionFactory
from hub.exports.energy_building_exports_factory import EnergyBuildingsExportsFactory
import pandas as pd
from geopandas import GeoDataFrame
from shapely.geometry import Polygon


class CityLayerTest(TestCase):

  @staticmethod
  def _prepare_buildings(bldgs_group):
    target_json = bldgs_group['target']
    adjacent_json = bldgs_group['adjacent']
    target_buildings = [f"building_{target_json['index']}"]
    adjacent_buildings = [f"building_{el}" for el in adjacent_json['Ids']]
    target_dict = [dict(
      name=f"building_{target_json['index']}",
      height=target_json['height_max'],
      idx=target_json['index'],
      uid=target_json['uid'],
      year_built=2005,
      # year_built= 2005 if target_json['year_built']==9999 else target_json['year_built'],
      coords=target_json['coords'],
      function="residential",
      # function= "residential" if target_json['year_built']>2000 else "industry",
    )]
    adjacent_dict = [dict(
      name=f"building_{el['index']}",
      height=el['height_max'],
      idx=el['index'],
      uid=el['uid'],
      year_built=2005,
      # year_built=2005 if el['year_built']==9999 else el['year_built'],
      coords=el['geom']['coordinates'],
      function="residential",
      # function= "residential" if el['year_built']>2000 else "industry",
    ) for el in adjacent_json['data']]
    df = pd.DataFrame(target_dict + adjacent_dict)
    geometries = [Polygon(row['coords'][0]) for ix, row in df.iterrows()]
    gdf = GeoDataFrame(df, crs="EPSG:4326", geometry=geometries)
    gdf = gdf.set_crs('EPSG:4326')
    gdf = gdf.to_crs('EPSG:26911')
    return gdf, target_buildings, adjacent_buildings

  def _genidf(self, bldgs_group):
    t0 = time.time()
    buildings_df, target_buildings, adjacent_buildings = self._prepare_buildings(bldgs_group)
    output_path = (Path(__file__).parent / 'tests_outputs').resolve()
    city = GeometryFactory('gpandas', data_frame=buildings_df).city
    ConstructionFactory('nrel', city).enrich()
    UsageFactory('comnet', city).enrich()
    EnergyBuildingsExportsFactory('idf', city, output_path, target_buildings=target_buildings,
                                  adjacent_buildings=adjacent_buildings).export_debug()
    filepath = os.path.join(output_path, city.name + ".idf")
    newfilepath = filepath[:-4] + "_" + uuid.uuid4().hex[:10] + ".idf"
    os.rename(filepath, newfilepath)
    print(f"It took {round((time.time() - t0), 0)} seconds")
    return newfilepath

  def test_city_layers(self):
    json_path = str((Path(__file__).parent / 'tests_data' / 'city_layers.json').resolve())
    with open(json_path) as json_file:
      data = json.loads(json_file.read())
    self._genidf(data)



"""
Occupancy module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from typing import Union, List
from hub.city_model_structure.attributes.schedule import Schedule


class Occupancy:
  """
  Occupancy class
  """
  def __init__(self):
    self._occupancy_density = None
    self._sensible_convective_internal_gain = None
    self._sensible_radiative_internal_gain = None
    self._latent_internal_gain = None
    self._occupancy_schedules = None
    self._occupants = None

  @property
  def occupancy_density(self) -> Union[None, float]:
    """
    Get density in persons per m2
    :return: None or float
    """
    return self._occupancy_density

  @occupancy_density.setter
  def occupancy_density(self, value):
    """
    Set density in persons per m2
    :param value: float
    """
    if value is not None:
      self._occupancy_density = float(value)

  @property
  def sensible_convective_internal_gain(self) -> Union[None, float]:
    """
    Get sensible convective internal gain in Watts per m2
    :return: None or float
    """
    return self._sensible_convective_internal_gain

  @sensible_convective_internal_gain.setter
  def sensible_convective_internal_gain(self, value):
    """
    Set sensible convective internal gain in Watts per m2
    :param value: float
    """
    if value is not None:
      self._sensible_convective_internal_gain = float(value)

  @property
  def sensible_radiative_internal_gain(self) -> Union[None, float]:
    """
    Get sensible radiant internal gain in Watts per m2
    :return: None or float
    """
    return self._sensible_radiative_internal_gain

  @sensible_radiative_internal_gain.setter
  def sensible_radiative_internal_gain(self, value):
    """
    Set sensible radiant internal gain in Watts per m2
    :param value: float
    """
    if value is not None:
      self._sensible_radiative_internal_gain = float(value)

  @property
  def latent_internal_gain(self) -> Union[None, float]:
    """
    Get latent internal gain in Watts per m2
    :return: None or float
    """
    return self._latent_internal_gain

  @latent_internal_gain.setter
  def latent_internal_gain(self, value):
    """
    Set latent internal gain in Watts per m2
    :param value: float
    """
    if value is not None:
      self._latent_internal_gain = float(value)

  @property
  def occupancy_schedules(self) -> Union[None, List[Schedule]]:
    """
    Get occupancy schedules
    dataType = fraction
    :return: None or [Schedule]
    """
    return self._occupancy_schedules

  @occupancy_schedules.setter
  def occupancy_schedules(self, value):
    """
    Set occupancy schedules
    dataType = fraction
    :param value: [Schedule]
    """
    self._occupancy_schedules = value



"""
DBFactory performs database create, delete and update operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project CoderPeter Yefi peteryefi@gmail.com
"""
from hub.city_model_structure.city import City
from hub.persistence import City as CityRepository
from hub.persistence import SimulationResults


class DBFactory:
  """
  DBFactory class
  """

  def __init__(self, db_name, dotenv_path, app_env):
    self._city_repository = CityRepository(db_name=db_name, dotenv_path=dotenv_path, app_env=app_env)
    self._simulation_results = SimulationResults(db_name=db_name, dotenv_path=dotenv_path, app_env=app_env)

  def persist_city(self, city: City, pickle_path, application_id: int, user_id: int):
    """
    Persist city into postgres database
    :param city: City to be stored
    :param pickle_path: Path to save the pickle file
    :param application_id: Application id owning this city
    :param user_id: User who create the city
    """
    return self._city_repository.insert(city, pickle_path, application_id, user_id)

  def update_city(self, city_id, city):
    """
   Update an existing city in postgres database
   :param city_id: the id of the city to update
   :param city: the updated city object
    """
    return self._city_repository.update(city_id, city)

  def delete_city(self, city_id):
    """
    Deletes a single city from postgres
    :param city_id: the id of the city to get
    """
    self._city_repository.delete(city_id)

  def add_simulation_results(self, name, values, city_id=None, city_object_id=None):
    """
    Add simulation results to the city or to the city_object
    :param name: simulation and simulation engine name
    :param values: simulation values in json format
    :param city_id: city id or None
    :param city_object_id: city object id or None
    """
    self._simulation_results.insert(name, values,city_id, city_object_id)



"""
CityGml module parses citygml_classes files and import the geometry into the city model structure
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Atiya atiya.atiya@mail.concordia.ca
"""
import xmltodict
from pathlib import Path
from hub.city_model_structure.lca_material import LcaMaterial as LMaterial


class LcaMaterial:
  def __init__(self, city, base_path):
    self._city = city
    self._base_path = base_path
    self._lca = None

  def enrich(self):
    self._city.lca_materials = []
    path = Path(self._base_path / 'lca_data.xml').resolve()

    with open(path) as xml:
      self._lca = xmltodict.parse(xml.read())

    for material in self._lca["library"]["building_materials"]['material']:
      _material = LMaterial()
      _material.type = material['@type']
      _material.id = material['@id']
      _material.name = material['@name']
      _material.density = material['density']['#text']
      _material.density_unit = material['density']['@unit']
      _material.embodied_carbon = material['embodied_carbon']['#text']
      _material.embodied_carbon_unit = material['embodied_carbon']['@unit']
      _material.recycling_ratio = material['recycling_ratio']
      _material.onsite_recycling_ratio = material['onsite_recycling_ratio']
      _material.company_recycling_ratio = material['company_recycling_ratio']
      _material.landfilling_ratio = material['landfilling_ratio']
      _material.cost = material['cost']['#text']
      _material._cost_unit = material['cost']['@unit']

      self._city.lca_materials.append(_material)



"""
Constant module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""


def validate_import_export_type(cls_name: type):
  """
  Retrieves all the function names in a class which are property types (decoration)
  and normal functions
  :param cls_name: the class name
  :return: [str], a list of functions in the class
  """
  return [func for func in dir(cls_name)
          if (type(getattr(cls_name, func)) is property or callable(getattr(cls_name, func)))
          and func in cls_name.__dict__ and func[0] == '_' and func != '__init__']



"""
User repository with database CRUD operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""

from hub.persistence import Repository
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy import select
from hub.persistence.models import User as Model
from hub.persistence.models import Application as ApplicationModel
from hub.persistence.models import UserRoles
from hub.helpers.auth import Auth
from typing import Union, Dict
from hub.hub_logger import logger
import datetime


class User(Repository):
  _instance = None

  def __init__(self, db_name: str, dotenv_path: str, app_env: str):
    super().__init__(db_name, dotenv_path, app_env)

  def __new__(cls, db_name, dotenv_path, app_env):
    """
    Implemented for a singleton pattern
    """
    if cls._instance is None:
      cls._instance = super(User, cls).__new__(cls)
    return cls._instance

  def insert(self, name: str, password: str, role: UserRoles, application_id: int) -> Union[Model, Dict]:
    """
    Inserts a new user
    :param name: user name
    :param password: user password
    :param role: user rol [Admin or Hub_Reader]
    :param application_id: user application id
    :return: [User, Dictionary]
    """
    user = self.get_by_name_and_application(name, application_id)
    if user is None:
      try:
        user = Model(name=name, password=Auth.hash_password(password), role=role, application_id=application_id)
        self.session.add(user)
        self.session.flush()
        self.session.commit()
        return user
      except SQLAlchemyError as err:
        logger.error(f'An error occurred while creating user: {err}')
    else:
      return {'message': f'user {name} already exists for that application'}

  def update(self, user_id: int, name: str, password: str, role: UserRoles) -> Union[Dict, None]:
    """
    Updates a user
    :param user_id: the id of the user to be updated
    :param name: the name of the user
    :param password: the password of the user
    :param role: the role of the user
    :return:
    """
    try:
      self.session.query(Model).filter(Model.id == user_id).update({
        'name': name,
        'password': Auth.hash_password(password),
        'role': role,
        'updated': datetime.datetime.utcnow()
      })
      self.session.commit()
    except SQLAlchemyError as err:
      logger.error(f'Error while updating user: {err}')
      return {'err_msg': 'Error occurred while updated user'}

  def delete(self, user_id: int):
    """
    Deletes a user with the id
    :param user_id: the user id
    :return: None
    """
    try:
      self.session.query(Model).filter(Model.id == user_id).delete()
      self.session.commit()
    except SQLAlchemyError as err:
      logger.error(f'Error while fetching user: {err}')

  def get_by_name_and_application(self, name: str, application_id: int) -> [Model]:
    """
    Fetch user based on the email address
    :param name: User name
    :param application_id: User application name
    :return: [User] matching the search criteria
    """
    try:
      return self.session.execute(
        select(Model).where(Model.name == name, Model.application_id == application_id)
      ).first()
    except SQLAlchemyError as err:
      logger.error(f'Error while fetching user by name and application: {err}')

  def get_by_name_application_id_and_password(self, name: str, password: str, application_id: int) -> [Model]:
    """
    Fetch user based on the email and password
    :param name: User name
    :param password: User password
    :param application_id: User password

    :return: [User]
    """
    try:
      user = self.session.execute(
        select(Model).where(Model.name == name, Model.application_id == application_id)
      ).first()
      if user:
        if Auth.check_password(password, user[0].password):
          #return user[0]
          return True
      return False
    except SQLAlchemyError as err:
      logger.error(f'Error while fetching user by email: {err}')

  def get_by_name_application_uuid_and_password(self, name: str, password: str, application_uuid: str) -> [Model]:
    """
    Fetch user based on the email and password
    :param name: User name
    :param password: User password
    :param application_uuid: Application uuid

    :return: [User]
    """
    try:
      application = self.session.execute(
        select(ApplicationModel).where(ApplicationModel.application_uuid == application_uuid)
      ).first()
      return self.get_by_name_application_id_and_password(name, password, application[0].id)
    except SQLAlchemyError as err:
      logger.error(f'Error while fetching user by name: {err}')



"""
Fast charging infrastructure module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class FastChargingInfrastructure:
  """
  FastChargingInfrastructure class
  """

  def __init__(self):
    self._electrical_demand = None
    self._losses_in_grid = None

  @property
  def electrical_demand(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._electrical_demand

  @property
  def losses_in_grid(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._losses_in_grid



"""
Test EnergySystemsFactory and various heatpump models
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
from unittest import TestCase
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.energy_systems_factory import EnergySystemsFactory
from hub.exports.energy_systems_factory import EnergySystemsExportFactory
from hub.imports.db_factory import DBFactory
from hub.exports.db_factory import DBFactory as ExportDBFactory
from hub.persistence.base_repo import BaseRepo
from sqlalchemy import create_engine
from hub.persistence.models import City
from hub.persistence.models import SimulationTypes
from hub.persistence.models import HeatPumpTypes
from hub.persistence.models import HeatPumpSimulation
from hub.persistence.models import User
from sqlalchemy.exc import ProgrammingError
from hub.imports.user_factory import UserFactory
from hub.persistence.models import UserRoles

# User defined paramenters
hp_sim_data = {
  'StartYear': 2020,
  'EndYear': 2021,
  'MaximumHPEnergyInput': 8000,
  'HoursOfStorageAtMaxDemand': 1,
  'BuildingSuppTemp': 40,
  'TemperatureDifference': 15,
  'FuelLHV': 47100,
  'FuelPrice': 0.12,
  'FuelEF': 1887,
  'FuelDensity': 0.717,
  'HPSupTemp': 60
}


class TestHeatPumpSimulation(TestCase):
  """
  Heat pump simulation test cases
  """
  @classmethod
  def setUpClass(cls) -> None:
    """
    Test setup
    :return: None
    """
    repo = BaseRepo(db_name='test_db', app_env='TEST', dotenv_path='../.env')
    eng = create_engine(f'postgresql://{repo.config.get_db_user()}@/{repo.config.get_db_user()}')

    try:
      conn = eng.connect()
      conn.execute('commit')
      conn.execute('DROP DATABASE test_db')
      conn.close()
    except ProgrammingError as err:
      print(f'Database does not exist. Nothing to delete')

    cnn = eng.connect()
    cnn.execute('commit')
    cnn.execute("CREATE DATABASE test_db")
    cnn.close()

    # Create test tables if they do not exit
    User.__table__.create(bind=repo.engine, checkfirst=True)
    City.__table__.create(bind=repo.engine, checkfirst=True)
    HeatPumpSimulation.__table__.create(bind=repo.engine, checkfirst=True)


    city_file = "tests_data/C40_Final.gml"
    cls._city = GeometryFactory('citygml', city_file).city
    EnergySystemsFactory('air source hp', cls._city).enrich()

    cls._db_factory = DBFactory(db_name='test_db', app_env='TEST', dotenv_path='../.env')
    cls._export_db_factory = ExportDBFactory(db_name='test_db', app_env='TEST', dotenv_path='../.env')
    user_factory = UserFactory(db_name='test_db', app_env='TEST', dotenv_path='../.env')
    cls._user = user_factory.create_user("Admin", "admin@hub.com", "Admin@123", UserRoles.Admin)

  def test_heat_pump_simulation_persistence(self):
    output = EnergySystemsExportFactory(city=self._city, user_input=hp_sim_data, hp_model='018',
                                        output_path=None, sim_type=1).export()
    hp_sim_data["HeatPumpModel"] = '018'
    hp_sim_data["SimulationType"] = SimulationTypes.Parallel
    hp_sim_data["HeatPumpType"] = HeatPumpTypes.Air
    hp_sim_data["HourlyElectricityDemand"] = output["hourly_electricity_demand"]
    hp_sim_data["DailyElectricityDemand"] = output["daily_electricity_demand"]
    hp_sim_data["MonthlyElectricityDemand"] = output["monthly_electricity_demand"]
    hp_sim_data["DailyFossilFuelConsumption"] = output["daily_fossil_consumption"]
    hp_sim_data["MonthlyFossilFuelConsumption"] = output["monthly_fossil_consumption"]

    saved_city = self._db_factory.persist_city(self._user.id, self._city)
    hp_sim = self._db_factory.persist_hp_simulation(hp_sim_data, saved_city.id)
    self.assertEqual(hp_sim.heat_pump_type, HeatPumpTypes.Air)
    self.assertEqual(hp_sim.simulation_type, SimulationTypes.Parallel)
    self.assertEqual(hp_sim.fuel_efficiency, hp_sim_data["FuelEF"])
    self.assertEqual(hp_sim.monthly_electricity_demand, output["monthly_electricity_demand"])
    self._db_factory.delete_hp_simulation(hp_sim.id)
    self._db_factory.delete_city(saved_city.id)

  def test_get_heat_pump_simulation_by_city(self):
    output = EnergySystemsExportFactory(city=self._city, user_input=hp_sim_data, hp_model='012',
                                        output_path=None, sim_type=0).export()
    hp_sim_data["HeatPumpModel"] = '012'
    hp_sim_data["SimulationType"] = SimulationTypes.Series
    hp_sim_data["HeatPumpType"] = HeatPumpTypes.Air
    hp_sim_data["HourlyElectricityDemand"] = output["hourly_electricity_demand"]
    hp_sim_data["DailyElectricityDemand"] = output["daily_electricity_demand"]
    hp_sim_data["MonthlyElectricityDemand"] = output["monthly_electricity_demand"]
    hp_sim_data["DailyFossilFuelConsumption"] = output["daily_fossil_consumption"]
    hp_sim_data["MonthlyFossilFuelConsumption"] = output["monthly_fossil_consumption"]

    saved_city = self._db_factory.persist_city(self._user.id, self._city)
    self._db_factory.persist_hp_simulation(hp_sim_data, saved_city.id)

    # retrieved saved simulation by city id
    hp_sim = self._export_db_factory.get_hp_simulation_by_city(saved_city.id)
    self.assertEqual(hp_sim[0].heat_pump_type, HeatPumpTypes.Air)
    self.assertEqual(hp_sim[0].simulation_type, SimulationTypes.Series)
    self.assertEqual(hp_sim[0].fuel_price, hp_sim_data["FuelPrice"])
    self.assertEqual(hp_sim[0].hourly_electricity_demand, output["hourly_electricity_demand"])
    self._db_factory.delete_hp_simulation(hp_sim[0].id)
    self._db_factory.delete_city(saved_city.id)




"""
water_to_water_hp module defines a water to water heat pump heat pump
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""

from typing import List
from hub.city_model_structure.energy_systems.heat_pump import HeatPump


class WaterToWaterHP(HeatPump):
  """
  WaterToWaterHP class
  """

  def __init__(self):
    super().__init__()
    self._entering_water_temp = None
    self._leaving_water_temp = None
    self._total_cooling_capacity = None
    self._power_demand = None
    self._flow_rate = None
    self._power_demand_coff = None  # a coefficients

  @property
  def entering_water_temp(self) -> List[float]:
    """
    Get entering water temperature in degree celsius
    :return: [[float]]
    """
    return self._entering_water_temp

  @entering_water_temp.setter
  def entering_water_temp(self, value):
    """
    Set entering water temperature in degree celsius
    :param value: [[float]]
    """
    if self._entering_water_temp is None:
      self._entering_water_temp = value

  @property
  def leaving_water_temp(self) -> List[float]:
    """
    Get leaving water temperature in degree celsius
    :return: [[float]]
    """
    return self._leaving_water_temp

  @leaving_water_temp.setter
  def leaving_water_temp(self, value):
    """
    Set the leaving water temperature in degree celsius
    :param value: [[float]]
    :return:
    """
    if self._leaving_water_temp is None:
      self._leaving_water_temp = value

  @property
  def total_cooling_capacity(self) -> List[float]:
    """
    Get total cooling capacity
    :return: [float]
    """
    return self._total_cooling_capacity

  @total_cooling_capacity.setter
  def total_cooling_capacity(self, value):
    """
    Set the value for total cooling capacity
    :param value: [float]
    :return:
    """
    if self._total_cooling_capacity is None:
      self._total_cooling_capacity = value

  @property
  def power_demand(self) -> List[float]:
    """
    Get power demand in kW
    :return: [float]
    """
    return self._power_demand

  @power_demand.setter
  def power_demand(self, value):
    """
    Set the value for power demand in kW
    :param value: [float]
    :return:
    """
    if self._power_demand is None:
      self._power_demand = value

  @property
  def flow_rate(self) -> List[float]:
    """
    Get flow rate in kg/s
    :return: [[float]]
    """
    return self._flow_rate

  @flow_rate.setter
  def flow_rate(self, value):
    """
    Set flow rate in kW
    :param value: [[float]]
    :return:
    """
    if self._flow_rate is None:
      self._flow_rate = value

  @property
  def power_demand_coff(self) -> List[float]:
    """
    Get power demand coefficients
    :return: [float]
    """
    return self._power_demand_coff

  @power_demand_coff.setter
  def power_demand_coff(self, value):
    """
    Set the value for power demand coefficients
    :param value: [float]
    :return:
    """
    if self._power_demand_coff is None:
      self._power_demand_coff = value



"""
User performs user related crud operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project CoderPeter Yefi peteryefi@gmail.com
"""
from hub.persistence import User

class UserFactory:
  """
  UserFactory class
  """

  def __init__(self, db_name, app_env, dotenv_path):
    self._user_repo = User(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)

  def login_user(self, email: str, password: str):
    """
    Retrieve a single city from postgres
    :param email: the email of the user
    :param password: the password of the user
    """
    return self._user_repo.get_user_by_email_and_password(email, password)

  def get_user_by_email(self, email):
    """
    Retrieve a single user
    :param email: the email of the user to get
    """
    return self._user_repo.get_by_email(email)



"""
Insel monthly energy balance
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guillermo.GutierrezMorote@concordia.ca
"""

from pathlib import Path
import pandas as pd
import csv
import hub.helpers.constants as cte

class InselMonthlyEnergyBalance:
  """
  Import SRA results
  """
  def __init__(self, city, base_path):

    self._city = city
    self._base_path = base_path

  @staticmethod
  def _demand(insel_output_file_path):
    heating = []
    cooling = []
    with open(Path(insel_output_file_path).resolve()) as csv_file:
      csv_reader = csv.reader(csv_file)
      for line in csv_reader:
        demand = str(line).replace("['", '').replace("']", '').split()
        for i in range(0, 2):
          if demand[i] != 'NaN':
            aux = float(demand[i]) * 1000  # kWh to Wh
            demand[i] = str(aux)
          else:
            demand[i] = '0'
        heating.append(demand[0])
        cooling.append(demand[1])
    monthly_heating = pd.DataFrame(heating, columns=[cte.INSEL_MEB])
    monthly_cooling = pd.DataFrame(cooling, columns=[cte.INSEL_MEB])
    return monthly_heating, monthly_cooling

  def enrich(self):
    for building in self._city.buildings:
      file_name = building.name + '.out'
      insel_output_file_path = Path(self._base_path / file_name).resolve()
      if insel_output_file_path.is_file():
        building.heating[cte.MONTH], building.cooling[cte.MONTH] = self._demand(insel_output_file_path)
        building.heating[cte.YEAR] = pd.DataFrame(
          [building.heating[cte.MONTH][cte.INSEL_MEB].sum()], columns=[cte.INSEL_MEB]
        )
        building.cooling[cte.YEAR] = pd.DataFrame(
          [building.cooling[cte.MONTH][cte.INSEL_MEB].sum()], columns=[cte.INSEL_MEB]
        )



"""
Lighting module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from typing import Union, List
from hub.city_model_structure.attributes.schedule import Schedule


class Lighting:
  """
  Lighting class
  """
  def __init__(self):
    self._density = None
    self._convective_fraction = None
    self._radiative_fraction = None
    self._latent_fraction = None
    self._schedules = None

  @property
  def density(self) -> Union[None, float]:
    """
    Get lighting density in Watts per m2
    :return: None or float
    """
    return self._density

  @density.setter
  def density(self, value):
    """
    Set lighting density in Watts per m2
    :param value: float
    """
    if value is not None:
      self._density = float(value)

  @property
  def convective_fraction(self) -> Union[None, float]:
    """
    Get convective fraction
    :return: None or float
    """
    return self._convective_fraction

  @convective_fraction.setter
  def convective_fraction(self, value):
    """
    Set convective fraction
    :param value: float
    """
    if value is not None:
      self._convective_fraction = float(value)

  @property
  def radiative_fraction(self) -> Union[None, float]:
    """
    Get radiant fraction
    :return: None or float
    """
    return self._radiative_fraction

  @radiative_fraction.setter
  def radiative_fraction(self, value):
    """
    Set radiant fraction
    :param value: float
    """
    if value is not None:
      self._radiative_fraction = float(value)

  @property
  def latent_fraction(self) -> Union[None, float]:
    """
    Get latent fraction
    :return: None or float
    """
    return self._latent_fraction

  @latent_fraction.setter
  def latent_fraction(self, value):
    """
    Set latent fraction
    :param value: float
    """
    if value is not None:
      self._latent_fraction = float(value)

  @property
  def schedules(self) -> Union[None, List[Schedule]]:
    """
    Get schedules
    dataType = fraction
    :return: None or [Schedule]
    """
    return self._schedules

  @schedules.setter
  def schedules(self, value):
    """
    Set schedules
    dataType = fraction
    :param value: [Schedule]
    """
    self._schedules = value



"""
SensorsFactory retrieve sensors information
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from pathlib import Path
from hub.hub_logger import logger
from hub.helpers.utils import validate_import_export_type


class SensorsFactory:
  """
  UsageFactory class
  """
  def __init__(self, handler, city, end_point, base_path=None):
    if base_path is None:
      base_path = Path(Path(__file__).parent.parent / 'data/sensors')
    self._handler = '_' + handler.lower().replace(' ', '_')
    class_funcs = validate_import_export_type(SensorsFactory)
    if self._handler not in class_funcs:
      err_msg = f"Wrong import type. Valid functions include {class_funcs}"
      logger.error(err_msg)
      raise Exception(err_msg)
    self._city = city
    self._end_point = end_point
    self._base_path = base_path

  def _cec(self):
    """
    Enrich the city by using concordia energy consumption sensors as data source
    """
    raise NotImplementedError('need to be reimplemented')

  def _cgf(self):
    """
    Enrich the city by using concordia gas flow sensors as data source
    """
    raise NotImplementedError('need to be reimplemented')

  def _ct(self):
    """
    Enrich the city by using concordia temperature sensors as data source
    """
    raise NotImplementedError('need to be reimplemented')

  def enrich(self):
    """
    Enrich the city given to the class using the given sensor handler
    :return: None
    """
    getattr(self, self._handler, lambda: None)()



"""
Geojson module parses geojson files and import the geometry into the city model structure
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guillermo Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""
import json

import trimesh.creation

from pyproj import Transformer
from shapely.geometry import Polygon as ShapelyPolygon

import hub.helpers.constants as cte
from hub.imports.geometry.helpers.geometry_helper import GeometryHelper
from hub.city_model_structure.attributes.polygon import Polygon
from hub.city_model_structure.building import Building
from hub.city_model_structure.building_demand.surface import Surface
from hub.city_model_structure.city import City


class Geojson:
  """
  Geojson class
  """
  X = 0
  Y = 1

  def __init__(self,
               path,
               extrusion_height_field=None,
               year_of_construction_field=None,
               function_field=None,
               function_to_hub=None):
    # todo: destination epsg should change according actual the location
    self._transformer = Transformer.from_crs('epsg:4326', 'epsg:26911')
    self._min_x = cte.MAX_FLOAT
    self._min_y = cte.MAX_FLOAT
    self._max_x = cte.MIN_FLOAT
    self._max_y = cte.MIN_FLOAT
    self._max_z = 0
    self._city = None
    self._extrusion_height_field = extrusion_height_field
    self._year_of_construction_field = year_of_construction_field
    self._function_field = function_field
    self._function_to_hub = function_to_hub
    with open(path) as json_file:
      self._geojson = json.loads(json_file.read())

  def _save_bounds(self, x, y):
    if x > self._max_x:
      self._max_x = x
    if x < self._min_x:
      self._min_x = x
    if y > self._max_y:
      self._max_y = y
    if y < self._min_y:
      self._min_y = y

  @staticmethod
  def _create_buildings_lod0(name, year_of_construction, function, surfaces_coordinates):
    surfaces = []
    buildings = []
    for zone, surface_coordinates in enumerate(surfaces_coordinates):
      points = GeometryHelper.points_from_string(GeometryHelper.remove_last_point_from_string(surface_coordinates))
      polygon = Polygon(points)
      surfaces.append(Surface(polygon, polygon))
      buildings.append(Building(f'{name}_zone_{zone}', surfaces, year_of_construction, function))
    return buildings

  @staticmethod
  def _create_buildings_lod1(name, year_of_construction, function, height, surface_coordinates):
    lod0_buildings = Geojson._create_buildings_lod0(name, year_of_construction, function, surface_coordinates)
    surfaces = []
    buildings = []
    for zone, lod0_building in enumerate(lod0_buildings):
      for surface in lod0_building.surfaces:
        shapely_polygon = ShapelyPolygon(surface.solid_polygon.coordinates)
        if not shapely_polygon.is_valid:
          print(surface.solid_polygon.area)
          print('error?', name, surface_coordinates)
          continue
        mesh = trimesh.creation.extrude_polygon(shapely_polygon, height)
        for face in mesh.faces:
          points = []
          for vertex_index in face:
            points.append(mesh.vertices[vertex_index])
          polygon = Polygon(points)
          surface = Surface(polygon, polygon)
          surfaces.append(surface)
        buildings.append(Building(f'{name}_zone_{zone}', surfaces, year_of_construction, function))
    return buildings

  def _get_polygons(self, polygons, coordinates):
    if type(coordinates[0][self.X]) != float:
      polygons = []
      for element in coordinates:
        polygons = self._get_polygons(polygons, element)
      return polygons
    else:
      transformed_coordinates = ''
      for coordinate in coordinates:
        transformed = self._transformer.transform(coordinate[self.Y], coordinate[self.X])
        self._save_bounds(transformed[self.X], transformed[self.Y])
        transformed_coordinates = f'{transformed_coordinates} {transformed[self.X]} {transformed[self.Y]} 0.0'
      polygons.append(transformed_coordinates.lstrip(' '))
      return polygons

  @property
  def city(self) -> City:
    """
    Get city out of a Geojson file
    """
    if self._city is None:
      missing_functions = []
      buildings = []
      building_id = 0
      for feature in self._geojson['features']:
        extrusion_height = 0
        if self._extrusion_height_field is not None:
          extrusion_height = float(feature['properties'][self._extrusion_height_field])
        year_of_construction = None
        if self._year_of_construction_field is not None:
          year_of_construction = int(feature['properties'][self._year_of_construction_field])
        function = None
        if self._function_field is not None:
          function = feature['properties'][self._function_field]
          if self._function_to_hub is not None:
            # use the transformation dictionary to retrieve the proper function
            if function in self._function_to_hub:
              function = self._function_to_hub[function]
            else:
              if function not in missing_functions:
                missing_functions.append(function)
              function = function
        geometry = feature['geometry']
        if 'id' in feature:
          building_name = feature['id']
        else:
          building_name = f'building_{building_id}'
          building_id += 1
        polygons = []
        lod = 1
        for part, coordinates in enumerate(geometry['coordinates']):
          polygons = self._get_polygons(polygons, coordinates)
          for zone, polygon in enumerate(polygons):
            if extrusion_height == 0:
              buildings = buildings + Geojson._create_buildings_lod0(f'{building_name}_part_{part}_zone{zone}',
                                                                     year_of_construction,
                                                                     function,
                                                                     [polygon])
              lod = 0
            else:
              if self._max_z < extrusion_height:
                self._max_z = extrusion_height
              buildings = buildings + Geojson._create_buildings_lod1(f'{building_name}_part_{part}',
                                                                     year_of_construction,
                                                                     function,
                                                                     extrusion_height,
                                                                     [polygon])

      self._city = City([self._min_x, self._min_y, 0.0], [self._max_x, self._max_y, self._max_z], 'epsg:26911')
      for building in buildings:
        self._city.add_city_object(building)
      self._city.level_of_detail.geometry = lod
      if len(missing_functions) > 0:
        print(f'There are unknown functions {missing_functions}')
    return self._city



"""
Level of detail module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""


class LevelOfDetail:
  """
  Level of detail for the city class
  """
  def __init__(self):
    self._geometry = None
    self._construction = None
    self._usage = None
    self._weather = None
    self._surface_radiation = None

  @property
  def geometry(self):
    """
    Get the city minimal geometry level of detail from 0 to 4
    :return: int
    """
    return self._geometry

  @geometry.setter
  def geometry(self, value):
    """
    Set the city minimal geometry level of detail from 0 to 4
    """
    self._geometry = value

  @property
  def construction(self):
    """
    Get the city minimal construction level of detail, 1 or 2
    :return: int
    """
    return self._construction

  @construction.setter
  def construction(self, value):
    """
    Set the city minimal construction level of detail, 1 or 2
    """
    self._construction = value

  @property
  def usage(self):
    """
    Get the city minimal usage level of detail, 1 or 2
    :return: int
    """
    return self._usage

  @usage.setter
  def usage(self, value):
    """
    Set the city minimal usage level of detail, 1 or 2
    """
    self._usage = value

  @property
  def weather(self):
    """
    Get the city minimal weather level of detail, 0 (yearly), 1 (monthly), 2 (hourly)
    :return: int
    """
    return self._weather

  @weather.setter
  def weather(self, value):
    """
    Set  the city minimal weather level of detail, 0 (yearly), 1 (monthly), 2 (hourly)
    """
    self._usage = value

  @property
  def surface_radiation(self):
    """
    Get the city minimal surface radiation level of detail, 0 (yearly), 1 (monthly), 2 (hourly)
    :return: int
    """
    return self._surface_radiation

  @surface_radiation.setter
  def surface_radiation(self, value):
    """
    Set  the city minimal surface radiation level of detail, 0 (yearly), 1 (monthly), 2 (hourly)
    """
    self._surface_radiation = value



"""
Simplified Radiosity Algorithm
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guillermo.GutierrezMorote@concordia.ca
"""
import xmltodict

from hub.imports.weather_factory import WeatherFactory
import hub.helpers.constants as cte


class SimplifiedRadiosityAlgorithm:
  """
  Export to SRA format
  """

  def __init__(self,
               city,
               file_name,
               weather_file,
               weather_format,
               target_buildings=None,
               begin_month=1,
               begin_day=1,
               end_month=12,
               end_day=31):
    self._file_name = file_name
    self._begin_month = begin_month
    self._begin_day = begin_day
    self._end_month = end_month
    self._end_day = end_day
    self._city = city
    self._target_buildings = target_buildings
    self._weather_format = weather_format
    self._weather_file = weather_file
    self._export()

  def _correct_point(self, point):
    # correct the x, y, z values into the reference frame set by city lower_corner
    x = point[0] - self._city.lower_corner[0]
    y = point[1] - self._city.lower_corner[1]
    z = point[2] - self._city.lower_corner[2]
    return [x, y, z]

  def _export(self):
    self._export_sra_xml()
    self._export_sra_cli()

  def _export_sra_cli(self):
    file = self._city.climate_file
    days_in_month = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    WeatherFactory(self._weather_format, self._city, file_name=self._weather_file).enrich()
    content = self._city.name + '\n'
    content += str(self._city.latitude) + ',' + str(self._city.longitude) + ',0.0,' + str(self._city.time_zone) + '\n'
    content += '\ndm  m h G_Dh  G_Bn\n'
    total_days = 0
    for month in range(1, 13):
      if month > 1:
        total_days += days_in_month[month - 2]
      for day in range(1, days_in_month[month - 1] + 1):
        for hour in range(1, 25):
          if month == 1:
            i = 24 * (day - 1) + hour - 1
          else:
            i = (total_days + day - 1) * 24 + hour - 1
          representative_building = self._city.buildings[0]
          content += str(day) + '  ' + str(month) + '  ' + str(hour) + '  ' \
                     + str(representative_building.global_horizontal[cte.HOUR].epw[i]) + '  ' \
                     + str(representative_building.beam[cte.HOUR].epw[i]) + '\n'
    with open(file, "w") as file:
      file.write(content)

  def _export_sra_xml(self):
    buildings = []
    for building_index, building in enumerate(self._city.buildings):
      if self._target_buildings is None:
        simulate = 'true'
      else:
        simulate = 'false'
        for target_building in self._target_buildings:
          if building.name == target_building.name:
            simulate = 'true'
      building_dict = {
        '@Name': f'{building.name}',
        '@id': f'{building_index}',
        '@key': f'{building.name}',
        '@Simulate': f'{simulate}'
      }
      walls, roofs, floors = [], [], []
      for surface in building.surfaces:
        surface_dict = {
          '@id': f'{surface.id}',
          '@ShortWaveReflectance': f'{surface.short_wave_reflectance}'
        }
        for point_index, point in enumerate(surface.perimeter_polygon.coordinates):
          point = self._correct_point(point)
          surface_dict[f'V{point_index}'] = {
            '@x': f'{point[0]}',
            '@y': f'{point[1]}',
            '@z': f'{point[2]}'
          }
        if surface.type == 'Wall':
          walls.append(surface_dict)
        elif surface.type == 'Roof':
          roofs.append(surface_dict)
        else:
          floors.append(surface_dict)
      building_dict['Wall'] = walls
      building_dict['Roof'] = roofs
      building_dict['Floor'] = floors
      buildings.append(building_dict)
    sra = {
      'CitySim': {
        '@name': f'{self._file_name.name}',
        'Simulation': {
          '@beginMonth': f'{self._begin_month}',
          '@beginDay': f'{self._begin_day}',
          '@endMonth': f'{self._end_month}',
          '@endDay': f'{self._end_day}',
        },
        'Climate': {
          '@location': f'{self._city.climate_file}',
          '@city': f'{self._city.climate_reference_city}'
        },
        'District': {
          'FarFieldObstructions': None,
          'Building': buildings
        }
      }
    }
    with open(self._file_name, "w") as file:
      file.write(xmltodict.unparse(sra, pretty=True, short_empty_elements=True))



"""
Test EnergySystemsFactory and various heatpump models
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
from unittest import TestCase
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.db_factory import DBFactory
from hub.imports.user_factory import UserFactory
from hub.exports.db_factory import DBFactory as ExportDBFactory
from hub.persistence.base_repo import BaseRepo
from sqlalchemy import create_engine
from hub.persistence.models import City
from hub.persistence.models import User, UserRoles
from pickle import loads
from sqlalchemy.exc import ProgrammingError


class TestDBFactory(TestCase):
  """
  TestDBFactory
  """

  @classmethod
  def setUpClass(cls) -> None:
    """
    Test setup
    :return: None
    """
    # Create test database
    repo = BaseRepo(db_name='test_db', app_env='TEST', dotenv_path='/usr/local/etc/hub/.env')
    eng = create_engine(f'postgresql://{repo.config.get_db_user()}@/{repo.config.get_db_user()}')

    try:
      # delete test database if it exists
      conn = eng.connect()
      conn.execute('commit')
      conn.execute('DROP DATABASE test_db')
      conn.close()
    except ProgrammingError as err:
      print(f'Database does not exist. Nothing to delete')

    cnn = eng.connect()
    cnn.execute('commit')
    cnn.execute("CREATE DATABASE test_db")
    cnn.close()
    User.__table__.create(bind=repo.engine, checkfirst=True)
    City.__table__.create(bind=repo.engine, checkfirst=True)

    city_file = "tests_data/C40_Final.gml"
    cls.city = GeometryFactory('citygml', city_file).city
    cls._db_factory = DBFactory(db_name='test_db', app_env='TEST', dotenv_path='../.env')
    cls._export_db_factory = ExportDBFactory(db_name='test_db', app_env='TEST', dotenv_path='../.env')
    user_factory = UserFactory(db_name='test_db', app_env='TEST', dotenv_path='../.env')
    cls._user = user_factory.create_user("Admin", "admin@hub.com", "Admin@123", UserRoles.Admin)

  def test_save_city(self):
    saved_city = self._db_factory.persist_city(self._user.id, self.city)
    self.assertEqual(saved_city.name, 'Montréal')
    pickled_city = loads(saved_city.city)
    self.assertEqual(len(pickled_city.buildings), 10)
    self.assertEqual(pickled_city.buildings[0].floor_area, 1990.9913970530033)
    self._db_factory.delete_city(saved_city.id)

  def test_save_same_city_with_same_hub_version(self):
    first_city = self._db_factory.persist_city(self._user.id, self.city)
    second_city = self._db_factory.persist_city(self._user.id, self.city)
    self.assertEqual(second_city['message'], f'Same version of {self.city.name} exist')
    self.assertEqual(first_city.name, 'Montréal')
    self.assertEqual(first_city.country_code, 'ca')
    self._db_factory.delete_city(first_city.id)

  def test_get_city_by_name(self):
    city = self._db_factory.persist_city(self._user.id, self.city)
    retrieved_city = self._export_db_factory.get_city_by_name(city.name)
    self.assertEqual(retrieved_city[0].lower_corner[0], 610610.7547462888)
    self._db_factory.delete_city(city.id)

  def test_get_city_by_user(self):
    city = self._db_factory.persist_city(self._user.id, self.city)
    retrieved_city = self._export_db_factory.get_city_by_user(self._user.id)
    self.assertEqual(retrieved_city[0].user_id, self._user.id)
    self._db_factory.delete_city(city.id)

  def test_get_city_by_id(self):
    city = self._db_factory.persist_city(self._user.id, self.city)
    retrieved_city = self._export_db_factory.get_city(city.id)
    self.assertEqual(retrieved_city.upper_corner[0], 610818.6731258357)
    self._db_factory.delete_city(city.id)

  def test_get_update_city(self):
    city = self._db_factory.persist_city(self._user.id, self.city)
    self.city.longitude = 1.43589
    self.city.latitude = -9.38928339
    self._db_factory.update_city(city.id, self.city)
    updated_city = self._export_db_factory.get_city(city.id)
    self.assertEqual(updated_city.longitude, 1.43589)
    self.assertEqual(updated_city.latitude, -9.38928339)
    self._db_factory.delete_city(city.id)



"""
Construction catalog content
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""


class Content:
  def __init__(self, archetypes, constructions, materials, windows):
    self._archetypes = archetypes
    self._constructions = constructions
    self._materials = materials
    self._windows = windows

  @property
  def archetypes(self):
    """
    All archetypes in the catalog
    """
    return self._archetypes

  @property
  def constructions(self):
    """
    All constructions in the catalog
    """
    return self._constructions

  @property
  def materials(self):
    """
    All materials in the catalog
    """
    return self._materials

  @property
  def windows(self):
    """
    All windows in the catalog
    """
    return self._windows



"""
export a city into Obj format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from pathlib import Path
import trimesh.exchange.obj
from hub.exports.formats.triangular import Triangular
from hub.imports.geometry_factory import GeometryFactory


class Obj(Triangular):
  """
  Export to obj format
  """
  def __init__(self, city, path):
    super().__init__(city, path, 'obj')

  def to_ground_points(self):
    """
    Move closer to the origin
    """
    file_name_in = self._city.name + '.' + self._triangular_format
    file_name_out = self._city.name + '_ground.' + self._triangular_format
    file_path_in = (Path(self._path).resolve() / file_name_in).resolve()
    file_path_out = (Path(self._path).resolve() / file_name_out).resolve()
    obj = GeometryFactory('obj', path=file_path_in)
    scene = obj.scene
    scene.rezero()
    obj_file = trimesh.exchange.obj.export_obj(scene)
    with open(file_path_out, 'w') as file:
      file.write(obj_file)



"""
Station
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
import uuid

from hub.city_model_structure.iot.sensor import Sensor


class Station:
  def __init__(self, station_id=None, _mobile=False):
    self._id = station_id
    self._mobile = _mobile
    self._sensors = []

  @property
  def id(self):
    """
    Get the station id a random uuid will be assigned if no ID was provided to the constructor
    :return: Id
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def mobile(self):
    """
    Get if the station is mobile or not
    :return: bool
    """
    return self._mobile

  @property
  def sensors(self) -> [Sensor]:
    """
    Get the sensors belonging to the station
    :return: [Sensor]
    """
    return self._sensors



"""
Obj module parses obj files and import the geometry into the city model structure
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import trimesh.exchange.load
from trimesh import Scene
import trimesh.geometry
from hub.city_model_structure.city import City
from hub.city_model_structure.building import Building
from hub.city_model_structure.building_demand.surface import Surface
from hub.city_model_structure.attributes.polygon import Polygon


class Obj:
  """
  Obj class
  """
  def __init__(self, path):
    self._city = None
    with open(path, 'r') as file:
      self._scene = trimesh.exchange.load.load(file, file_type='obj', force='scene')
    self._corners = self._scene.bounds_corners
    _bound_corner_min = None
    _bound_corner_max = None
    for corner in self._corners:
      if _bound_corner_min is None:
        _bound_corner_min = corner
      elif _bound_corner_max is None:
        _bound_corner_max = corner
      else:
        _bound_corner_min[0] = min(_bound_corner_min[0], corner[0])
        _bound_corner_min[1] = min(_bound_corner_min[1], corner[1])
        _bound_corner_min[2] = min(_bound_corner_min[2], corner[2])
        _bound_corner_max[0] = max(_bound_corner_max[0], corner[0])
        _bound_corner_max[1] = max(_bound_corner_max[1], corner[1])
        _bound_corner_max[2] = max(_bound_corner_max[2], corner[2])
    self._lower_corner = _bound_corner_min
    self._upper_corner = _bound_corner_max

  @property
  def scene(self) -> Scene:
    """
    Get obj scene
    """
    return self._scene

  @property
  def city(self) -> City:
    """
    Get city out of an obj file
    """
    lod = 0
    if self._city is None:
      # todo: refactor this method to clearly choose the obj type
      # todo: where do we get this information from?
      srs_name = 'EPSG:26911'

      self._city = City(self._lower_corner, self._upper_corner, srs_name)
      scene = self.scene.geometry
      keys = scene.keys()
      for key in keys:
        name = key
        # todo: where do we get this information from?
        lod = 1
        year_of_construction = 0
        function = ''

        obj = scene[key]
        surfaces = []
        for face in obj.faces:
          # todo: review for obj with windows
          points = []
          for vertex_index in face:
            points.append(obj.vertices[vertex_index])
          solid_polygon = Polygon(points)
          perimeter_polygon = solid_polygon
          surface = Surface(solid_polygon, perimeter_polygon)
          surfaces.append(surface)
        building = Building(name, surfaces, year_of_construction, function, terrains=None)
        self._city.add_city_object(building)
      self._city.level_of_detail.geometry = lod
    return self._city



"""
NrcanPhysicsParameters import the construction and material information defined by NRCAN
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import sys

from hub.catalog_factories.construction_catalog_factory import ConstructionCatalogFactory
from hub.city_model_structure.building_demand.layer import Layer
from hub.city_model_structure.building_demand.material import Material
from hub.helpers.dictionaries import Dictionaries
from hub.imports.construction.helpers.construction_helper import ConstructionHelper
from hub.imports.construction.helpers.storeys_generation import StoreysGeneration


class NrcanPhysicsParameters:
  """
  NrcanPhysicsParameters class
  """
  def __init__(self, city, base_path, divide_in_storeys=False):
    self._city = city
    self._path = base_path
    self._divide_in_storeys = divide_in_storeys
    self._climate_zone = ConstructionHelper.city_to_nrel_climate_zone(city.name)

  def enrich_buildings(self):
    """
    Returns the city with the construction parameters assigned to the buildings
    """
    city = self._city
    canel_catalog = ConstructionCatalogFactory('nrcan').catalog
    for building in city.buildings:
      try:
        function = Dictionaries().hub_function_to_nrcan_construction_function[building.function]
        archetype = self._search_archetype(canel_catalog, function, building.year_of_construction,
                                           self._climate_zone)
      except KeyError:
        sys.stderr.write(f'Building {building.name} has unknown construction archetype for building function: '
                         f'{building.function} and building year of construction: {building.year_of_construction} '
                         f'and climate zone reference norm {self._climate_zone}\n')
        return

      # if building has no thermal zones defined from geometry, and the building will be divided in storeys,
      # one thermal zone per storey is assigned
      if len(building.internal_zones) == 1:
        if building.internal_zones[0].thermal_zones is None:
          self._create_storeys(building, archetype, self._divide_in_storeys)
        if self._divide_in_storeys:
          for internal_zone in building.internal_zones:
            for thermal_zone in internal_zone.thermal_zones:
              thermal_zone.total_floor_area = thermal_zone.footprint_area
        else:
          number_of_storeys = int(float(building.eave_height) / float(building.average_storey_height))
          thermal_zone = building.internal_zones[0].thermal_zones[0]
          thermal_zone.total_floor_area = thermal_zone.footprint_area * number_of_storeys
      else:
        for internal_zone in building.internal_zones:
          for thermal_zone in internal_zone.thermal_zones:
            thermal_zone.total_floor_area = thermal_zone.footprint_area

      for internal_zone in building.internal_zones:
        self._assign_values(internal_zone.thermal_zones, archetype)
        for thermal_zone in internal_zone.thermal_zones:
          self._calculate_view_factors(thermal_zone)

  @staticmethod
  def _search_archetype(nrel_catalog, function, year_of_construction, climate_zone):
    nrel_archetypes = nrel_catalog.entries('archetypes')
    for building_archetype in nrel_archetypes:
      construction_period_limits = building_archetype.construction_period.split(' - ')
      if construction_period_limits[1] == 'PRESENT':
        construction_period_limits[1] = 3000
      if int(construction_period_limits[0]) <= int(year_of_construction) < int(construction_period_limits[1]):
        if (str(function) == str(building_archetype.function)) and \
           (climate_zone == str(building_archetype.climate_zone)):
          return building_archetype
    raise KeyError('archetype not found')

  @staticmethod
  def _search_construction_in_archetype(archetype, construction_type):
    construction_archetypes = archetype.constructions
    for construction_archetype in construction_archetypes:
      if str(construction_type) == str(construction_archetype.type):
        return construction_archetype
    return None

  def _assign_values(self, thermal_zones, archetype):
    for thermal_zone in thermal_zones:
      thermal_zone.additional_thermal_bridge_u_value = archetype.extra_loses_due_to_thermal_bridges
      thermal_zone.effective_thermal_capacity = archetype.thermal_capacity
      thermal_zone.indirectly_heated_area_ratio = archetype.indirect_heated_ratio
      thermal_zone.infiltration_rate_system_on = archetype.infiltration_rate_for_ventilation_system_on
      thermal_zone.infiltration_rate_system_off = archetype.infiltration_rate_for_ventilation_system_off
      for thermal_boundary in thermal_zone.thermal_boundaries:
        construction_archetype = self._search_construction_in_archetype(archetype, thermal_boundary.type)
        thermal_boundary.construction_name = construction_archetype.name
        try:
          thermal_boundary.window_ratio = construction_archetype.window_ratio
        except ValueError:
          # This is the normal operation way when the windows are defined in the geometry
          continue
        thermal_boundary.layers = []
        for layer_archetype in construction_archetype.layers:
          layer = Layer()
          layer.thickness = layer_archetype.thickness
          material = Material()
          archetype_material = layer_archetype.material
          material.name = archetype_material.name
          material.id = archetype_material.id
          material.no_mass = archetype_material.no_mass
          if archetype_material.no_mass:
            material.thermal_resistance = archetype_material.thermal_resistance
          else:
            material.density = archetype_material.density
            material.conductivity = archetype_material.conductivity
            material.specific_heat = archetype_material.specific_heat
          material.solar_absorptance = archetype_material.solar_absorptance
          material.thermal_absorptance = archetype_material.thermal_absorptance
          material.visible_absorptance = archetype_material.visible_absorptance
          layer.material = material
          thermal_boundary.layers.append(layer)
        # The agreement is that the layers are defined from outside to inside
        external_layer = construction_archetype.layers[0]
        external_surface = thermal_boundary.parent_surface
        external_surface.short_wave_reflectance = 1 - float(external_layer.material.solar_absorptance)
        external_surface.long_wave_emittance = 1 - float(external_layer.material.solar_absorptance)
        internal_layer = construction_archetype.layers[len(construction_archetype.layers) - 1]
        internal_surface = thermal_boundary.internal_surface
        internal_surface.short_wave_reflectance = 1 - float(internal_layer.material.solar_absorptance)
        internal_surface.long_wave_emittance = 1 - float(internal_layer.material.solar_absorptance)

        for thermal_opening in thermal_boundary.thermal_openings:
          if construction_archetype.window is not None:
            window_archetype = construction_archetype.window
            thermal_opening.construction_name = window_archetype.name
            thermal_opening.frame_ratio = window_archetype.frame_ratio
            thermal_opening.g_value = window_archetype.g_value
            thermal_opening.overall_u_value = window_archetype.overall_u_value

  # todo: verify windows
  @staticmethod
  def _calculate_view_factors(thermal_zone):
    """
    Get thermal zone view factors matrix
    :return: [[float]]
    """
    total_area = 0
    for thermal_boundary in thermal_zone.thermal_boundaries:
      total_area += thermal_boundary.opaque_area
      for thermal_opening in thermal_boundary.thermal_openings:
        total_area += thermal_opening.area

    view_factors_matrix = []
    for thermal_boundary_1 in thermal_zone.thermal_boundaries:
      values = []
      for thermal_boundary_2 in thermal_zone.thermal_boundaries:
        value = 0
        if thermal_boundary_1.id != thermal_boundary_2.id:
          value = thermal_boundary_2.opaque_area / (total_area - thermal_boundary_1.opaque_area)
        values.append(value)
      for thermal_boundary in thermal_zone.thermal_boundaries:
        for thermal_opening in thermal_boundary.thermal_openings:
          value = thermal_opening.area / (total_area - thermal_boundary_1.opaque_area)
          values.append(value)
      view_factors_matrix.append(values)

    for thermal_boundary_1 in thermal_zone.thermal_boundaries:
      values = []
      for thermal_opening_1 in thermal_boundary_1.thermal_openings:
        for thermal_boundary_2 in thermal_zone.thermal_boundaries:
          value = thermal_boundary_2.opaque_area / (total_area - thermal_opening_1.area)
          values.append(value)
        for thermal_boundary in thermal_zone.thermal_boundaries:
          for thermal_opening_2 in thermal_boundary.thermal_openings:
            value = 0
            if thermal_opening_1.id != thermal_opening_2.id:
              value = thermal_opening_2.area / (total_area - thermal_opening_1.area)
            values.append(value)
        view_factors_matrix.append(values)
    thermal_zone.view_factors_matrix = view_factors_matrix

  @staticmethod
  def _create_storeys(building, archetype, divide_in_storeys):
    building.average_storey_height = archetype.average_storey_height
    building.storeys_above_ground = 1
    thermal_zones = StoreysGeneration(building, building.internal_zones[0],
                                      divide_in_storeys=divide_in_storeys).thermal_zones
    building.internal_zones[0].thermal_zones = thermal_zones



"""
EnergySystemsFactory exports energy systems into several formats
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""

from pathlib import Path
from hub.exports.energy_systems.air_source_hp_export import AirSourceHPExport
from hub.exports.energy_systems.water_to_water_hp_export import WaterToWaterHPExport


class EnergySystemsExportFactory:
  """
  Exports factory class for energy systems
  """

  def __init__(self, city, user_input, hp_model, output_path, sim_type=0, data_type='heat', base_path=None,
               demand_path=None):
    """

    :param city: the city object
    :param user_input: user provided input from UI
    :param hp_model: the heat pump model to run
    :param output_path: the file to hold simulation results
    :param sim_type: the simulation type, 0 for series 1 for parallel
    :param data_type: indicates whether cooling or heating data is used
    :param base_path: the data directory of energy systems
    :param demand_path: path to hourly energy dempand file
    """

    self._city = city
    if base_path is None:
      base_path = Path(Path(__file__).parent.parent / 'data/energy_systems')
    self._base_path = base_path
    self._user_input = user_input
    self._hp_model = hp_model
    self._data_type = data_type
    self._output_path = output_path
    self._sim_type = sim_type
    self._demand_path = demand_path

  def _export_heat_pump(self, source):
    """
    Exports heat pump performance data as coefficients
    of some objective function
    :return: None
    """
    if source == 'air':
      return AirSourceHPExport(self._base_path, self._city, self._output_path, self._sim_type, self._demand_path)\
        .execute_insel(self._user_input, self._hp_model, self._data_type)
    elif source == 'water':
      return WaterToWaterHPExport(self._base_path, self._city, self._output_path, self._sim_type, self._demand_path)\
        .execute_insel(self._user_input, self._hp_model)

  def export(self, source='air'):
    """
    Export the city given to the class using the given export type handler
    :return: None
    """
    return getattr(self, '_export_heat_pump', lambda: None)(source)



"""
DBFactory performs read related operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project CoderPeter Yefi peteryefi@gmail.com
"""
from hub.persistence import City
from hub.persistence import Application
from hub.persistence import User
from hub.persistence import CityObject


class DBFactory:
  """
  DBFactory class
  """

  def __init__(self, db_name, app_env, dotenv_path):
    self._city = City(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)
    self._application = Application(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)
    self._user = User(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)
    self._city_object = CityObject(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)

  def get_city(self, city_id):
    """
    Retrieve a single city from postgres
    :param city_id: the id of the city to get
    """
    return self._city.get_by_id(city_id)

  def get_city_by_name(self, city_name):
    """
    Retrieve a single city from postgres
    :param city_name: the name of the city to get
    """
    return self._city.get_by_name(city_name)

  def get_city_by_user(self, user_id):
    """
    Retrieve cities created by user
    :param user_id: the id of the user
    """
    return self._city.get_by_user(user_id)

  def application_info(self, application_uuid):
    return self._application.get_by_uuid(application_uuid)

  def user_info(self, name, password, application_id):
    return self._user.get_by_name_application_id_and_password(name, password, application_id)

  def user_login(self, name, password, application_uuid):
    return self._user.get_by_name_application_uuid_and_password(name, password, application_uuid)

  def building_info(self, name, city_id):
    return self._city_object.get_by_name_and_city(name, city_id)




import glob
import pathlib
from distutils.util import convert_path

import pkg_resources
from setuptools import setup

with pathlib.Path('hub/requirements.txt').open() as r:
    install_requires = [
        str(requirement)
        for requirement
        in pkg_resources.parse_requirements(r)
    ]

install_requires.append('setuptools')

main_ns = {}
version = convert_path('hub/version.py')
with open(version) as f:
    exec(f.read(), main_ns)


setup(
    name='cerc-hub',
    version=main_ns['__version__'],
    description="CERC Hub consist in a set of classes (Central data model), importers and exporters to help researchers "
                "to create better and sustainable cities",
    long_description="CERC Hub consist in a set of classes (Central data model), importers and exporters to help "
                     "researchers to create better and sustainable cities.\n\nDevelop at Concordia university in canada "
                     "as part of the research group from the next generation cities institute our aim among others it's "
                     "to provide a comprehensive set of tools to help researchers and urban developers to make decisions "
                     "to improve the livability and efficiency of our cities",
    classifiers=[
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    include_package_data=True,
    packages=['hub',
              'hub.catalog_factories',
              'hub.catalog_factories.construction',
              'hub.catalog_factories.data_models',
              'hub.catalog_factories.data_models.construction',
              'hub.catalog_factories.data_models.greenery',
              'hub.catalog_factories.data_models.usages',
              'hub.catalog_factories.greenery',
              'hub.catalog_factories.greenery.ecore_greenery',
              'hub.catalog_factories.usage',
              'hub.city_model_structure',
              'hub.city_model_structure.attributes',
              'hub.city_model_structure.building_demand',
              'hub.city_model_structure.energy_systems',
              'hub.city_model_structure.greenery',
              'hub.city_model_structure.iot',
              'hub.city_model_structure.transport',
              'hub.config',
              'hub.data',
              'hub.exports',
              'hub.exports.building_energy',
              'hub.exports.building_energy.idf_files',
              'hub.exports.building_energy.insel',
              'hub.exports.energy_systems',
              'hub.exports.formats',
              'hub.helpers',
              'hub.helpers.data',
              'hub.hub_logger',
              'hub.imports',
              'hub.imports.construction',
              'hub.imports.construction.helpers',
              #'hub.imports.construction.data_classes',
              'hub.imports.energy_systems',
              'hub.imports.geometry',
              'hub.imports.geometry.citygml_classes',
              'hub.imports.geometry.helpers',
              'hub.imports.results',
              'hub.imports.usage',
              'hub.imports.weather',
              'hub.imports.weather.helpers',
              'hub.persistence',
              'hub.persistence.models',
              'hub.persistence.repositories',
              'hub.imports'
              ],
    setup_requires=install_requires,
    data_files=[
        ('hub', glob.glob('hub/requirements.txt')),
        ('hub/config', glob.glob('hub/config/*.ini')),
        ('hub/catalog_factories/greenery/ecore_greenery', glob.glob('hub/catalog_factories/greenery/ecore_greenery/*.ecore')),
        ('hub/data/construction.', glob.glob('hub/data/construction/*')),
        ('hub/data/customized_imports', glob.glob('hub/data/customized_imports/*.xml')),
        ('hub/data/energy_systems', glob.glob('hub/data/energy_systems/*.xml')),
        ('hub/data/energy_systems', glob.glob('hub/data/energy_systems/*.insel')),
        ('hub/data/energy_systems', glob.glob('hub/data/energy_systems/*.xlsx')),
        ('hub/data/energy_systems', glob.glob('hub/data/energy_systems/*.txt')),
        ('hub/data/energy_systems', glob.glob('hub/data/energy_systems/*.yaml')),
        ('hub/data/greenery', glob.glob('hub/data/greenery/*.xml')),
        ('hub/data/life_cycle_assessment', glob.glob('hub/data/life_cycle_assessment/*.xml')),
        ('hub/data/schedules', glob.glob('hub/data/schedules/*.xml')),
        ('hub/data/schedules', glob.glob('hub/data/schedules/*.xlsx')),
        ('hub/data/schedules/idf_files', glob.glob('hub/data/schedules/idf_files/*.idf')),
        ('hub/data/sensors', glob.glob('hub/data/sensors/*.json')),
        ('hub/data/usage', glob.glob('hub/data/usage/*.xml')),
        ('hub/data/usage', glob.glob('hub/data/usage/*.xlsx')),
        ('hub/data/weather', glob.glob('hub/data/weather/*.dat')),
        ('hub/data/weather/epw', glob.glob('hub/data/weather/epw/*.epw')),
        ('hub/data/weather', glob.glob('hub/data/weather/*.dat')),
        ('hub/exports/building_energy/idf_files', glob.glob('hub/exports/building_energy/idf_files/*.idf')),
        ('hub/exports/building_energy/idf_files', glob.glob('hub/exports/building_energy/idf_files/*.idd')),
        ('hub/helpers/data', glob.glob('hub/helpers/data/quebec_to_hub.json'))
    ],

)



"""
CityGmlLod1 module parses citygml_classes files with level of detail 1 and import the geometry into the city model
structure
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from hub.imports.geometry.helpers.geometry_helper import GeometryHelper
from hub.imports.geometry.citygml_classes.citygml_base import CityGmlBase
from hub.city_model_structure.building_demand.surface import Surface
from hub.city_model_structure.attributes.polygon import Polygon


class CityGmlLod1(CityGmlBase):
  """
  CityGmlLod1 class to parse level of detail 1 city gml files
  """

  @classmethod
  def _multi_curve(cls, city_object_member):
    pass

  def __init__(self, city_object_member):
    super().__init__()
    self._city_object_member = city_object_member
    self._surfaces = self._identify(self._city_object_member)

  @classmethod
  def _identify(cls, city_object_member):
    if 'lod1Solid' in city_object_member:
      return cls._solid(city_object_member)
    if 'lod1MultiSurface' in city_object_member:
      return cls._multi_surface(city_object_member)
    raise NotImplementedError(city_object_member)

  @classmethod
  def _solid(cls, city_object_member):
    try:
      solid_points = [
        GeometryHelper.points_from_string(GeometryHelper.remove_last_point_from_string(
          s['Polygon']['exterior']['LinearRing']['posList']['#text']))
        for s in city_object_member['lod1Solid']['Solid']['exterior']['CompositeSurface']['surfaceMember']]
    except TypeError:
      solid_points = [
        GeometryHelper.points_from_string(GeometryHelper.remove_last_point_from_string(
          s['Polygon']['exterior']['LinearRing']['posList']))
        for s in city_object_member['lod1Solid']['Solid']['exterior']['CompositeSurface']['surfaceMember']]

    return [Surface(Polygon(sp), Polygon(sp)) for sp in solid_points]

  @classmethod
  def _multi_surface(cls, city_object_member):
    solid_points = [GeometryHelper.points_from_string(GeometryHelper.remove_last_point_from_string(
      s['Polygon']['exterior']['LinearRing']['posList']))
                    for s in city_object_member['Building']['lod1MultiSurface']['MultiSurface']['surfaceMember']]
    return [Surface(Polygon(sp), Polygon(sp)) for sp in solid_points]



"""
InternalGain module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from typing import Union, List
from hub.city_model_structure.attributes.schedule import Schedule


class InternalGain:
  """
  InternalGain class
  """

  def __init__(self):
    self._type = None
    self._average_internal_gain = None
    self._convective_fraction = None
    self._radiative_fraction = None
    self._latent_fraction = None
    self._schedules = None

  @property
  def type(self) -> Union[None, str]:
    """
    Get internal gains type
    :return: None or string
    """
    return self._type

  @type.setter
  def type(self, value):
    """
    Set internal gains type
    :param value: string
    """
    if value is not None:
      self._type = str(value)

  @property
  def average_internal_gain(self) -> Union[None, float]:
    """
    Get internal gains average internal gain in W/m2
    :return:  None or float
    """
    return self._average_internal_gain

  @average_internal_gain.setter
  def average_internal_gain(self, value):
    """
    Set internal gains average internal gain in W/m2
    :param value: float
    """
    if value is not None:
      self._average_internal_gain = float(value)

  @property
  def convective_fraction(self) -> Union[None, float]:
    """
    Get internal gains convective fraction
    :return:  None or float
    """
    return self._convective_fraction

  @convective_fraction.setter
  def convective_fraction(self, value):
    """
    Set internal gains convective fraction
    :param value: float
    """
    if value is not None:
      self._convective_fraction = float(value)

  @property
  def radiative_fraction(self) -> Union[None, float]:
    """
    Get internal gains radiative fraction
    :return:  None or float
    """
    return self._radiative_fraction

  @radiative_fraction.setter
  def radiative_fraction(self, value):
    """
    Set internal gains convective fraction
    :param value: float
    """
    if value is not None:
      self._radiative_fraction = float(value)

  @property
  def latent_fraction(self) -> Union[None, float]:
    """
    Get internal gains latent fraction
    :return: None or float
    """
    return self._latent_fraction

  @latent_fraction.setter
  def latent_fraction(self, value):
    """
    Set internal gains latent fraction
    :param value: float
    """
    if value is not None:
      self._latent_fraction = float(value)

  @property
  def schedules(self) -> Union[None, List[Schedule]]:
    """
    Get internal gain schedules
    data type = any number
    time step = 1 hour
    time range = 1 day
    :return: [Schedule]
    """
    return self._schedules

  @schedules.setter
  def schedules(self, value):
    """
    Set internal gain schedules
    data type = any number
    time step = 1 hour
    time range = 1 day
    :param value: [Schedule]
    """
    self._schedules = value



"""
weather helper
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import math
import hub.helpers.constants as cte
import pandas as pd
import calendar as cal
import numpy as np


class Weather:
  """
  Weather class
  """

  @staticmethod
  def sky_temperature(ambient_temperature):
    """
    Get sky temperature from ambient temperature in Celsius
    :return: float
    """
    # Swinbank - Source sky model approximation(1963) based on cloudiness statistics(32 %) in United States
    # ambient temperatures( in °C)
    # sky temperatures( in °C)
    values = []
    for temperature in ambient_temperature:
      value = 0.037536 * math.pow((temperature + cte.KELVIN), 1.5) \
              + 0.32 * (temperature + cte.KELVIN) - cte.KELVIN
      values.append(value)
    return values

  def get_monthly_mean_values(self, values):
    out = None
    if values is not None:
      if 'month' not in values.columns:
        values = pd.concat([self.month_hour, pd.DataFrame(values)], axis=1)
      out = values.groupby('month', as_index=False).mean()
      del out['month']
    return out

  def get_yearly_mean_values(self, values):
    return values.mean()

  def get_total_month(self, values):
    out = None
    if values is not None:
      if 'month' not in values.columns:
        values = pd.concat([self.month_hour, pd.DataFrame(values)], axis=1)
      out = pd.DataFrame(values).groupby('month', as_index=False).sum()
      del out['month']
    return out

  @property
  def month_hour(self):
    """
    returns a DataFrame that has x values of the month number (January = 1, February = 2...),
    being x the number of hours of the corresponding month
    :return: DataFrame(int)
    """
    array = []
    for i in range(0, 12):
      days_of_month = cal.monthrange(2015, i+1)[1]
      total_hours = days_of_month * 24
      array = np.concatenate((array, np.full(total_hours, i + 1)))
    return pd.DataFrame(array, columns=['month'])



"""
ThermalBoundary module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import uuid
from typing import List, Union, TypeVar
from hub.helpers.configuration_helper import ConfigurationHelper as ch
import hub.helpers.constants as cte
from hub.city_model_structure.building_demand.layer import Layer
from hub.city_model_structure.building_demand.thermal_opening import ThermalOpening
from hub.city_model_structure.building_demand.thermal_zone import ThermalZone

Surface = TypeVar('Surface')


class ThermalBoundary:
  """
  ThermalBoundary class
  """
  def __init__(self, parent_surface, opaque_area, windows_areas):
    self._parent_surface = parent_surface
    self._opaque_area = opaque_area
    self._windows_areas = windows_areas
    self._id = None
    self._thermal_zones = None
    self._thermal_openings = None
    self._layers = None
    self._he = ch().convective_heat_transfer_coefficient_exterior
    self._hi = ch().convective_heat_transfer_coefficient_interior
    self._u_value = None
    self._construction_name = None
    self._thickness = None
    self._internal_surface = None
    self._window_ratio = None
    self._window_ratio_is_calculated = False

  @property
  def id(self):
    """
    Get thermal zone id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def parent_surface(self) -> Surface:
    """
    Get the surface that belongs to the thermal boundary
    :return: Surface
    """
    return self._parent_surface

  @property
  def thermal_zones(self) -> List[ThermalZone]:
    """
    Get the thermal zones delimited by the thermal boundary
    :return: [ThermalZone]
    """
    return self._thermal_zones

  @thermal_zones.setter
  def thermal_zones(self, value):
    """
    Get the thermal zones delimited by the thermal boundary
    :param value: [ThermalZone]
    """
    self._thermal_zones = value

  @property
  def opaque_area(self):
    """
    Get the thermal boundary area in square meters
    :return: float
    """
    return float(self._opaque_area)

  @property
  def thickness(self):
    """
    Get the thermal boundary thickness in meters
    :return: float
    """
    if self._thickness is None:
      self._thickness = 0.0
      if self.layers is not None:
        for layer in self.layers:
          if not layer.material.no_mass:
            self._thickness += layer.thickness
    return self._thickness

  @property
  def thermal_openings(self) -> Union[None, List[ThermalOpening]]:
    """
    Get thermal boundary thermal openings
    :return: None or [ThermalOpening]
    """
    if self._thermal_openings is None:
      if self.window_ratio is not None:
        if self.window_ratio == 0:
          self._thermal_openings = []
        else:
          thermal_opening = ThermalOpening()
          if self.window_ratio == 1:
            _area = self.opaque_area
          else:
            _area = self.opaque_area * self.window_ratio / (1-self.window_ratio)
          thermal_opening.area = _area
          self._thermal_openings = [thermal_opening]
      else:
        if len(self.windows_areas) > 0:
          self._thermal_openings = []
          for window_area in self.windows_areas:
            thermal_opening = ThermalOpening()
            thermal_opening.area = window_area
            self._thermal_openings.append(thermal_opening)
        else:
          self._thermal_openings = []
    return self._thermal_openings

  @property
  def construction_name(self) -> Union[None, str]:
    """
    Get construction name
    :return: None or str
    """
    return self._construction_name

  @construction_name.setter
  def construction_name(self, value):
    """
    Set construction name
    :param value: str
    """
    if value is not None:
      self._construction_name = str(value)

  @property
  def layers(self) -> List[Layer]:
    """
    Get thermal boundary layers
    :return: [Layers]
    """
    return self._layers

  @layers.setter
  def layers(self, value):
    """
    Set thermal boundary layers
    :param value: [Layer]
    """
    self._layers = value

  @property
  def type(self):
    """
    Get thermal boundary surface type
    :return: str
    """
    return self.parent_surface.type

  @property
  def window_ratio(self) -> Union[None, float]:
    """
    Get thermal boundary window ratio
    It returns the window ratio calculated as the total windows' areas in a wall divided by
    the total (opaque + transparent) area of that wall if windows are defined in the geometry imported.
    If not, it returns the window ratio imported from an external source (e.g. construction library, manually assigned).
    If none of those sources are available, it returns None.
    :return: float
    """
    if self.windows_areas is not None:
      if not self._window_ratio_is_calculated:
        _calculated = True
        if len(self.windows_areas) == 0:
          self._window_ratio = 0
        else:
          total_window_area = 0
          for window_area in self.windows_areas:
            total_window_area += window_area
          self._window_ratio = total_window_area / (self.opaque_area + total_window_area)
    return self._window_ratio

  @window_ratio.setter
  def window_ratio(self, value):
    """
    Set thermal boundary window ratio
    :param value: str
    """
    if self._window_ratio_is_calculated:
      raise ValueError('Window ratio cannot be assigned when the windows are defined in the geometry.')
    self._window_ratio = float(value)

  @property
  def windows_areas(self) -> [float]:
    """
    Get windows areas
    :return: [float]
    """
    return self._windows_areas

  @property
  def u_value(self) -> Union[None, float]:
    """
    Get thermal boundary U-value in W/m2K
    internal and external convective coefficient in W/m2K values, can be configured at configuration.ini
    :return: None or float
    """
    if self._u_value is None:
      h_i = self.hi
      h_e = self.he
      if self.type == cte.GROUND:
        r_value = 1.0 / h_i + ch().soil_thickness / ch().soil_conductivity
      else:
        r_value = 1.0/h_i + 1.0/h_e
      try:
        for layer in self.layers:
          if layer.material.no_mass:
            r_value += float(layer.material.thermal_resistance)
          else:
            r_value += float(layer.thickness) / float(layer.material.conductivity)
        self._u_value = 1.0/r_value
      except TypeError:
        raise Exception('Constructions layers are not initialized') from TypeError
    return self._u_value

  @u_value.setter
  def u_value(self, value):
    """
    Set thermal boundary U-value in W/m2K
    :param value: float
    """
    if value is not None:
      self._u_value = float(value)

  @property
  def hi(self) -> Union[None, float]:
    """
    Get internal convective heat transfer coefficient (W/m2K)
    :return: None or float
    """
    return self._hi

  @hi.setter
  def hi(self, value):
    """
    Set internal convective heat transfer coefficient (W/m2K)
    :param value: float
    """
    if value is not None:
      self._hi = value

  @property
  def he(self) -> Union[None, float]:
    """
    Get external convective heat transfer coefficient (W/m2K)
    :return: None or float
    """
    return self._he

  @he.setter
  def he(self, value):
    """
    Set external convective heat transfer coefficient (W/m2K)
    :param value: float
    """
    if value is not None:
      self._he = value

  @property
  def internal_surface(self) -> Surface:
    """
    Get the internal surface of the thermal boundary
    :return: Surface
    """
    if self._internal_surface is None:
      self._internal_surface = self.parent_surface.inverse
    return self._internal_surface



"""
Test EnergySystemsFactory and various heatpump models
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
from unittest import TestCase
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.energy_systems_factory import EnergySystemsFactory
from hub.city_model_structure.energy_systems.water_to_water_hp import WaterToWaterHP
from hub.exports.energy_systems_factory import EnergySystemsExportFactory
import pandas as pd

import os

# User defined paramenters
user_input = {
  'StartYear': 2020,
  'EndYear': 2021,
  'MaximumHPEnergyInput': 8000,
  'HoursOfStorageAtMaxDemand': 1,
  'BuildingSuppTemp': 40,
  'TemperatureDifference': 15,
  'FuelLHV': 47100,
  'FuelPrice': 0.12,
  'FuelEF': 1887,
  'FuelDensity': 0.717,
  'HPSupTemp': 60
}


class TestEnergySystemsFactory(TestCase):
  """
  TestEnergySystemsFactory for Water to Water HP
  """

  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    city_file = "tests_data/C40_Final.gml"
    self._output_path = "tests_data/w2w_user_output.csv"
    self._city = GeometryFactory('citygml', path=city_file).city
    EnergySystemsFactory('water to water hp', self._city).enrich()

  def test_water_to_water_heat_pump_import(self):
    self.assertIsNotNone(self._city.energy_systems, 'City has energy systems')
    self.assertIsInstance(self._city.energy_systems[0].water_to_water_hp, WaterToWaterHP)
    self.assertEqual(self._city.energy_systems[0].water_to_water_hp.model, 'ClimateMaster 156 kW')
    self.assertEqual(self._city.energy_systems[2].water_to_water_hp.model, 'ClimateMaster 335 kW')

  def test_water_to_water_heat_pump_export(self):
    # User defined paramenters
    user_input = {
      'StartYear': 2020,
      'EndYear': 2021,
      'MaximumHPEnergyInput': 8000,
      'HoursOfStorageAtMaxDemand': 1,
      'BuildingSuppTemp': 40,
      'TemperatureDifference': 15,
      'FuelLHV': 47100,
      'FuelPrice': 0.12,
      'FuelEF': 1887,
      'FuelDensity': 0.717,
      'HPSupTemp': 60,
      'b1': 10,
      'b2': 10,
      'b3': 10,
      'b4': 10,
      'b5': 10,
      'b6': 10,
      'b7': 10,
      'b8': 10,
      'b9': 10,
      'b10': 10,
      'b11': 10
    }

    EnergySystemsExportFactory(city=self._city, user_input=user_input, hp_model='ClimateMaster 256 kW',
                               output_path=self._output_path, sim_type=1).export('water')
    df = pd.read_csv(self._output_path)
    self.assertEqual(df.shape, (13, 3))
    self.assertEqual(df.iloc[0, 1], 1031544.62)

  def tearDown(self) -> None:
    try:
      os.remove(self._output_path)
    except OSError:
      pass



"""
Soil class
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class Soil:
  def __init__(self, name, roughness, dry_conductivity, dry_density, dry_specific_heat, thermal_absorptance,
               solar_absorptance, visible_absorptance, saturation_volumetric_moisture_content,
               residual_volumetric_moisture_content):
    self._name = name
    self._roughness = roughness
    self._dry_conductivity = dry_conductivity
    self._dry_density = dry_density
    self._dry_specific_heat = dry_specific_heat
    self._thermal_absorptance = thermal_absorptance
    self._solar_absorptance = solar_absorptance
    self._visible_absorptance = visible_absorptance
    self._saturation_volumetric_moisture_content = saturation_volumetric_moisture_content
    self._residual_volumetric_moisture_content = residual_volumetric_moisture_content
    self._initial_volumetric_moisture_content = None

  @property
  def name(self):
    """
    Get soil name
    :return: string
    """
    return self._name

  @property
  def roughness(self):
    """
    Get soil roughness
    :return: string
    """
    return self._roughness

  @property
  def dry_conductivity(self):
    """
    Get soil dry conductivity in W/mK
    :return: float
    """
    return self._dry_conductivity

  @property
  def dry_density(self):
    """
    Get soil dry density in kg/m3
    :return: float
    """
    return self._dry_density

  @property
  def dry_specific_heat(self):
    """
    Get soil dry specific heat in J/kgK
    :return: float
    """
    return self._dry_specific_heat

  @property
  def thermal_absorptance(self):
    """
    Get soil thermal absortance
    :return: float
    """
    return self._thermal_absorptance

  @property
  def solar_absorptance(self):
    """
    Get soil solar absortance
    :return: float
    """
    return self._solar_absorptance

  @property
  def visible_absorptance(self):
    """
    Get soil visible absortance
    :return: float
    """
    return self._visible_absorptance

  @property
  def saturation_volumetric_moisture_content(self):
    """
    Get soil saturation volumetric moisture content
    :return: float
    """
    return self._saturation_volumetric_moisture_content

  @property
  def residual_volumetric_moisture_content(self):
    """
    Get soil residual volumetric moisture content
    :return: None or float
    """
    return self._residual_volumetric_moisture_content

  @residual_volumetric_moisture_content.setter
  def residual_volumetric_moisture_content(self, value):
    """
    Set soil residual volumetric moisture content
    :param value: float
    """
    self._residual_volumetric_moisture_content = value

  @property
  def initial_volumetric_moisture_content(self):
    """
    Get soil initial volumetric moisture content
    :return: None or float
    """
    return self._initial_volumetric_moisture_content

  @initial_volumetric_moisture_content.setter
  def initial_volumetric_moisture_content(self, value):
    """
    Set soil initial volumetric moisture content
    :param value: float
    """
    self._initial_volumetric_moisture_content = value



"""
Construction catalog construction
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
from hub.catalog_factories.data_models.construction.layer import Layer
from hub.catalog_factories.data_models.construction.window import Window


class Construction:
  def __init__(self, construction_id, construction_type, name, layers, window_ratio=None, window=None):
    self._id = construction_id
    self._type = construction_type
    self._name = name
    self._layers = layers
    self._window_ratio = window_ratio
    self._window = window

  @property
  def id(self):
    """
    Get construction id
    :return: int
    """
    return self._id

  @property
  def type(self):
    """
    Get construction type
    :return: str
    """
    return self._type

  @property
  def name(self):
    """
    Get construction name
    :return: str
    """
    return self._name

  @property
  def layers(self) -> [Layer]:
    """
    Get construction layers
    :return: [layer]
    """
    return self._layers

  @property
  def window_ratio(self):
    """
    Get construction window ratio
    :return: float
    """
    return self._window_ratio

  @property
  def window(self) -> Window:
    """
    Get construction window
    :return: Window
    """
    return self._window




"""
CityObject module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from typing import List, Union

from hub.city_model_structure.iot.sensor import Sensor
from hub.city_model_structure.building_demand.surface import Surface
from hub.city_model_structure.attributes.polyhedron import Polyhedron
from hub.helpers.configuration_helper import ConfigurationHelper


class CityObject:
  """
  class CityObject
  """
  def __init__(self, name, surfaces):
    self._name = name
    self._surfaces = surfaces
    self._type = None
    self._city_object_lower_corner = None
    self._detailed_polyhedron = None
    self._simplified_polyhedron = None
    self._min_x = ConfigurationHelper().max_coordinate
    self._min_y = ConfigurationHelper().max_coordinate
    self._min_z = ConfigurationHelper().max_coordinate
    self._centroid = None
    self._external_temperature = dict()
    self._global_horizontal = dict()
    self._diffuse = dict()
    self._beam = dict()
    self._sensors = []

  @property
  def name(self):
    """
    Get city object name
    :return: str
    """
    return self._name

  @property
  def type(self) -> str:
    """
    Get city object type
    :return: str
    """
    return self._type

  @property
  def volume(self) -> float:
    """
    Get city object volume in cubic meters
    :return: float
    """
    return self.simplified_polyhedron.volume

  @property
  def detailed_polyhedron(self) -> Polyhedron:
    """
    Get city object polyhedron including details such as holes
    :return: Polyhedron
    """
    if self._detailed_polyhedron is None:
      polygons = []
      for surface in self.surfaces:
        polygons.append(surface.solid_polygon)
        if surface.holes_polygons is not None:
          for hole_polygon in surface.holes_polygons:
            polygons.append(hole_polygon)
      self._detailed_polyhedron = Polyhedron(polygons)
    return self._detailed_polyhedron

  @property
  def simplified_polyhedron(self) -> Polyhedron:
    """
    Get city object polyhedron, just the simple lod2 representation
    :return: Polyhedron
    """
    if self._simplified_polyhedron is None:
      polygons = []
      for surface in self.surfaces:
        polygons.append(surface.perimeter_polygon)
      self._simplified_polyhedron = Polyhedron(polygons)
    return self._simplified_polyhedron

  @property
  def surfaces(self) -> List[Surface]:
    """
    Get city object surfaces
    :return: [Surface]
    """
    return self._surfaces

  def surface(self, name) -> Union[Surface, None]:
    """
    Get the city object surface with a given name
    :param name: str
    :return: None or Surface
    """
    for s in self.surfaces:
      if s.name == name:
        return s
    return None

  def surface_by_id(self, identification_number) -> Union[Surface, None]:
    """
    Get the city object surface with a given name
    :param identification_number: str
    :return: None or Surface
    """
    for s in self.surfaces:
      if str(s.id) == str(identification_number):
        return s
    return None

  @property
  def centroid(self) -> List[float]:
    """
    Get city object centroid
    :return: [x,y,z]
    """
    if self._centroid is None:
      self._centroid = self.simplified_polyhedron.centroid
    return self._centroid

  @property
  def max_height(self) -> float:
    """
    Get city object maximal height in meters
    :return: float
    """
    return self.simplified_polyhedron.max_z

  @property
  def external_temperature(self) -> {float}:
    """
    Get external temperature surrounding the city object in Celsius
    :return: dict{DataFrame(float)}
    """
    return self._external_temperature

  @external_temperature.setter
  def external_temperature(self, value):
    """
    Set external temperature surrounding the city object in Celsius
    :param value: dict{DataFrame(float)}
    """
    self._external_temperature = value

  @property
  def global_horizontal(self) -> dict:
    """
    Get global horizontal radiation surrounding the city object in W/m2
    :return: dict{DataFrame(float)}
    """
    return self._global_horizontal

  @global_horizontal.setter
  def global_horizontal(self, value):
    """
    Set global horizontal radiation surrounding the city object in W/m2
    :param value: dict{DataFrame(float)}
    """
    self._global_horizontal = value

  @property
  def diffuse(self) -> dict:
    """
    Get diffuse radiation surrounding the city object in W/m2
    :return: dict{DataFrame(float)}
    """
    return self._diffuse

  @diffuse.setter
  def diffuse(self, value):
    """
    Set diffuse radiation surrounding the city object in W/m2
    :param value: dict{DataFrame(float)}
    """
    self._diffuse = value

  @property
  def beam(self) -> dict:
    """
    Get beam radiation surrounding the city object in W/m2
    :return: dict{DataFrame(float)}
    """
    return self._beam

  @beam.setter
  def beam(self, value):
    """
    Set beam radiation surrounding the city object in W/m2
    :param value: dict{DataFrame(float)}
    """
    self._beam = value

  @property
  def lower_corner(self):
    """
    Get city object lower corner coordinates [x, y, z]
    :return: [x,y,z]
    """
    if self._city_object_lower_corner is None:
      self._city_object_lower_corner = [self._min_x, self._min_y, self._min_z]
    return self._city_object_lower_corner

  @property
  def sensors(self) -> List[Sensor]:
    """
    Get sensors belonging to the city object
    :return: [Sensor]
    """
    return self._sensors

  @sensors.setter
  def sensors(self, value):
    """
    Set sensors belonging to the city object
    :param value: [Sensor]
    """
    self._sensors = value



"""
City Object repository with database CRUD operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import datetime
from typing import Union, Dict

from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

from hub.hub_logger import logger
from hub.persistence import Repository
from hub.persistence.models import CityObject as Model
from hub.city_model_structure.building import Building


class CityObject(Repository):
  _instance = None

  def __init__(self, db_name: str, dotenv_path: str, app_env: str):
    super().__init__(db_name, dotenv_path, app_env)

  def __new__(cls, db_name, dotenv_path, app_env):
    """
    Implemented for a singleton pattern
    """
    if cls._instance is None:
      cls._instance = super(CityObject, cls).__new__(cls)
    return cls._instance

  def insert(self, city_id:int, building: Building) -> Union[Model, Dict]:
    """
    Inserts a new city object
    :param city_id: city id for the city owning this city object
    :param building: the city object (only building for now) to be inserted
    return CityObject model and dictionary
    """
    city_object = self.get_by_name_and_city(building.name, city_id)
    if city_object is None:
      try:
        object_usage = ''
        for internal_zone in building.internal_zones:
          for usage in internal_zone.usages:
            object_usage = f'{object_usage}{usage.name}_{usage.percentage} '
        object_usage = object_usage.rstrip()
        city_object = Model(city_id=city_id,
                            name=building.name,
                            alias=building.alias,
                            object_type=building.type,
                            year_of_construction=building.year_of_construction,
                            function=building.function,
                            usage=object_usage,
                            volume=building.volume,
                            area=building.floor_area)
        self.session.add(city_object)
        self.session.flush()
        self.session.commit()
        return city_object
      except SQLAlchemyError as err:
        logger.error(f'An error occurred while creating city_object: {err}')
    else:
      return {'message': f'A city_object named {building.name} already exists in that city'}

  def update(self,city_id: int, building: Building) -> Union[Dict, None]:
    """
    Updates an application
    :param city_id: the city id of the city owning the city object
    :param building: the city object
    :return:
    """
    try:
      object_usage = ''
      for internal_zone in building.internal_zones:
        for usage in internal_zone.usages:
          object_usage = f'{object_usage}{usage.name}_{usage.percentage} '
      object_usage = object_usage.rstrip()
      self.session.query(Model).filter(Model.name == building.name, Model.city_id == city_id).update(
        {'name': building.name,
         'alias': building.alias,
         'object_type': building.type,
         'year_of_construction': building.year_of_construction,
         'function': building.function,
         'usage': object_usage,
         'volume': building.volume,
         'area': building.floor_area,
         'updated': datetime.datetime.utcnow()})
      self.session.commit()
    except SQLAlchemyError as err:
      logger.error(f'Error while updating city object: {err}')
      return {'message': 'Error occurred while updating application'}

  def delete(self, city_id: int, name: str):
    """
    Deletes an application with the application_uuid
    :param city_id: The id for the city owning the city object
    :param name: The city object name
    :return: None
    """
    try:
      self.session.query(Model).filter(Model.city_id == city_id, Model.name == name).delete()
      self.session.commit()
    except SQLAlchemyError as err:
      logger.error(f'Error while deleting application: {err}')

  def get_by_name_and_city(self, name, city_id) -> [Model]:
    """
    Fetch a city object based on name and city id
    :param name: city object name
    :param city_id: a city identifier
    :return: [CityObject] with the provided name belonging to the city with id city_id
    """
    try:
      _city_object = self.session.execute(select(Model).where(
        Model.name == name, Model.city_id == city_id
      )).first()
      if _city_object is None:
        return None
      return _city_object[0]
    except SQLAlchemyError as err:
      print(err)
      logger.error(f'Error while fetching application by application_uuid: {err}')



"""
Test EnergySystemsFactory and various heatpump models
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
import pandas as pd
from unittest import TestCase
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.energy_systems_factory import EnergySystemsFactory
from hub.city_model_structure.energy_systems.air_source_hp import AirSourceHP
from hub.exports.energy_systems_factory import EnergySystemsExportFactory
import os

# User defined paramenters
user_input = {
  'StartYear': 2020,
  'EndYear': 2021,
  'MaximumHPEnergyInput': 8000,
  'HoursOfStorageAtMaxDemand': 1,
  'BuildingSuppTemp': 40,
  'TemperatureDifference': 15,
  'FuelLHV': 47100,
  'FuelPrice': 0.12,
  'FuelEF': 1887,
  'FuelDensity': 0.717,
  'HPSupTemp': 60
}


class TestEnergySystemsFactory(TestCase):
  """
  TestBuilding TestCase 1
  """

  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    city_file = "tests_data/C40_Final.gml"
    self._output_path = "tests_data/as_user_output.csv"
    self._city = GeometryFactory('citygml', path=city_file).city
    EnergySystemsFactory('air source hp', self._city).enrich()

  def test_air_source_heat_pump_import(self):
    self.assertIsNotNone(self._city.energy_systems, 'City has energy systems')
    self.assertIsInstance(self._city.energy_systems[0].air_source_hp, AirSourceHP)
    self.assertEqual(self._city.energy_systems[0].air_source_hp.model, '012')
    self.assertEqual(self._city.energy_systems[16].air_source_hp.model, '140')

  def test_air_source_series_heat_pump_export(self):
    EnergySystemsExportFactory(city=self._city, user_input=user_input, hp_model='012',
                               output_path=self._output_path).export()
    df = pd.read_csv(self._output_path)
    self.assertEqual(df.shape, (13, 3))
    self.assertEqual(df.iloc[0, 1], 1867715.88)

  def test_air_source_parallel_heat_pump_export(self):
    output = EnergySystemsExportFactory(city=self._city, user_input=user_input, hp_model='018',
                                        output_path=None, sim_type=1).export()
    self.assertEqual(output["hourly_electricity_demand"][0], 38748.5625)
    self.assertIsNotNone(output["daily_fossil_consumption"])
    self.assertEqual(len(output["hourly_electricity_demand"]), 8760)

  def tearDown(self) -> None:
    try:
      os.remove(self._output_path)
    except OSError:
      pass



"""
UsageZone module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
import uuid
from typing import Union, List
import hub.helpers.constants as cte
from hub.city_model_structure.building_demand.occupancy import Occupancy
from hub.city_model_structure.building_demand.lighting import Lighting
from hub.city_model_structure.building_demand.appliances import Appliances
from hub.city_model_structure.building_demand.thermal_control import ThermalControl
from hub.city_model_structure.building_demand.internal_gain import InternalGain


class UsageZone:
  """
  UsageZone class
  """
  def __init__(self):
    self._id = None
    self._usage = None
    self._percentage = None
    self._internal_gains = None
    self._hours_day = None
    self._days_year = None
    self._mechanical_air_change = None
    self._occupancy = None
    self._lighting = None
    self._appliances = None
    self._thermal_control = None

  @property
  def id(self):
    """
    Get usage zone id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def usage(self) -> Union[None, str]:
    """
    Get usage zone usage
    :return: None or str
    """
    return self._usage

  @usage.setter
  def usage(self, value):
    """
    Set usage zone usage
    :param value: str
    """
    if value is not None:
      self._usage = str(value)

  @property
  def percentage(self):
    """
    Get usage zone percentage in range[0,1]
    :return: float
    """
    return self._percentage

  @percentage.setter
  def percentage(self, value):
    """
    Set usage zone percentage in range[0,1]
    :param value: float
    """
    if value is not None:
      self._percentage = float(value)

  @property
  def internal_gains(self) -> List[InternalGain]:
    """
    Calculates and returns the list of all internal gains defined
    :return: InternalGains
    """
    if self._internal_gains is None:
      if self.occupancy is not None:
        if self.occupancy.latent_internal_gain is not None:
          _internal_gain = InternalGain()
          _internal_gain.type = cte.OCCUPANCY
          _total_heat_gain = (self.occupancy.sensible_convective_internal_gain
                              + self.occupancy.sensible_radiative_internal_gain
                              + self.occupancy.latent_internal_gain)
          _internal_gain.average_internal_gain = _total_heat_gain
          _internal_gain.latent_fraction = self.occupancy.latent_internal_gain / _total_heat_gain
          _internal_gain.radiative_fraction = self.occupancy.sensible_radiative_internal_gain / _total_heat_gain
          _internal_gain.convective_fraction = self.occupancy.sensible_convective_internal_gain / _total_heat_gain
          _internal_gain.schedules = self.occupancy.occupancy_schedules
          self._internal_gains = [_internal_gain]
      if self.lighting is not None:
        _internal_gain = InternalGain()
        _internal_gain.type = cte.LIGHTING
        _internal_gain.average_internal_gain = self.lighting.density
        _internal_gain.latent_fraction = self.lighting.latent_fraction
        _internal_gain.radiative_fraction = self.lighting.radiative_fraction
        _internal_gain.convective_fraction = self.lighting.convective_fraction
        _internal_gain.schedules = self.lighting.schedules
        if self._internal_gains is not None:
          self._internal_gains.append(_internal_gain)
        else:
          self._internal_gains = [_internal_gain]
      if self.appliances is not None:
        _internal_gain = InternalGain()
        _internal_gain.type = cte.APPLIANCES
        _internal_gain.average_internal_gain = self.appliances.density
        _internal_gain.latent_fraction = self.appliances.latent_fraction
        _internal_gain.radiative_fraction = self.appliances.radiative_fraction
        _internal_gain.convective_fraction = self.appliances.convective_fraction
        _internal_gain.schedules = self.appliances.schedules
        if self._internal_gains is not None:
          self._internal_gains.append(_internal_gain)
        else:
          self._internal_gains = [_internal_gain]
    return self._internal_gains

  @internal_gains.setter
  def internal_gains(self, value):
    """
    Set usage zone internal gains
    :param value: [InternalGain]
    """
    self._internal_gains = value

  @property
  def hours_day(self) -> Union[None, float]:
    """
    Get usage zone usage hours per day
    :return: None or float
    """
    return self._hours_day

  @hours_day.setter
  def hours_day(self, value):
    """
    Set usage zone usage hours per day
    :param value: float
    """
    if value is not None:
      self._hours_day = float(value)

  @property
  def days_year(self) -> Union[None, float]:
    """
    Get usage zone usage days per year
    :return: None or float
    """
    return self._days_year

  @days_year.setter
  def days_year(self, value):
    """
    Set usage zone usage days per year
    :param value: float
    """
    if value is not None:
      self._days_year = float(value)

  @property
  def mechanical_air_change(self) -> Union[None, float]:
    """
    Get usage zone mechanical air change in air change per hour (ACH)
    :return: None or float
    """
    return self._mechanical_air_change

  @mechanical_air_change.setter
  def mechanical_air_change(self, value):
    """
    Set usage zone mechanical air change in air change per hour (ACH)
    :param value: float
    """
    if value is not None:
      self._mechanical_air_change = float(value)

  @property
  def occupancy(self) -> Union[None, Occupancy]:
    """
    Get occupancy in the usage zone
    :return: None or Occupancy
    """
    return self._occupancy

  @occupancy.setter
  def occupancy(self, value):
    """
    Set occupancy in the usage zone
    :param value: Occupancy
    """
    self._occupancy = value

  @property
  def lighting(self) -> Union[None, Lighting]:
    """
    Get lighting information
    :return: None or Lighting
    """
    return self._lighting

  @lighting.setter
  def lighting(self, value):
    """
    Set lighting information
    :param value: Lighting
    """
    self._lighting = value

  @property
  def appliances(self) -> Union[None, Appliances]:
    """
    Get appliances information
    :return: None or Appliances
    """
    return self._appliances

  @appliances.setter
  def appliances(self, value):
    """
    Set appliances information
    :param value: Appliances
    """
    self._appliances = value

  @property
  def thermal_control(self) -> Union[None, ThermalControl]:
    """
    Get thermal control of this thermal zone
    :return: None or ThermalControl
    """
    return self._thermal_control

  @thermal_control.setter
  def thermal_control(self, value):
    """
    Set thermal control for this thermal zone
    :param value: ThermalControl
    """
    self._thermal_control = value
